from datetime import datetime

import morepath
import requests
import sqlalchemy
from more.jinja2 import Jinja2App
from morepath_sqlalchemy.app import Session
from morepath_sqlalchemy.model import Base
from sqlalchemy import INTEGER, Column, String, select, insert
from webob.exc import HTTPNotFound, HTTPServerError
from pybrake.middleware.morepath import init_app


engine = sqlalchemy.create_engine("sqlite:///:memory:")
Session.configure(bind=engine)

city_list = ["pune", "austin", "santabarbara", "washington"]


class User(Base):
    __tablename__ = "user"

    id = Column(INTEGER, primary_key=True)
    username = Column(String(80), unique=True, nullable=False)
    email = Column(String(120), unique=True, nullable=False)

    def __repr__(self):
        return "<User %r>" % self.username


class App(Jinja2App):
    pass


@App.template_directory()
def get_template_directory():
    return "templates"


class Weather(object):
    def __init__(self, data):
        self.data = data


@App.path(path='')
class Root(object):
    pass


@App.path(path='/locations')
class Locations(object):
    pass


@App.path(path='/date')
class Date(object):
    pass


@App.json(model=Weather)
def weather_default(self, request):
    return self.data


@App.path(path='weather/{location_name}', model=Weather)
def weather_data(location_name):
    if location_name not in city_list:
        raise HTTPNotFound("Location not found!")

    with requests.get(
            'https://airbrake.github.io/weatherapi/weather/' + location_name) as f:
        return Weather(f.json())


@App.json(model=Locations)
def locations(self, request):
    return {"locations": city_list}


@App.json(model=Date)
def date(self, request):
    return {"date": "Current date and time is: %s" % datetime.now()}


@App.html(model=Root, template="index.jinja2")
def hello_doc(self, request):
    with engine.begin() as conn:
        stmt = select(User)
        result = conn.execute(stmt)
        res = result.scalar()
        if not res:
            stmt = insert(User).values(username="test",
                                       email="test@test.com")
            conn.execute(stmt)
    return {}


if __name__ == '__main__':
    App.init_settings({
        "PYBRAKE": {
            "project_id": 99999,
            "project_key": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        }
    })
    Base.metadata.create_all(engine)
    morepath.commit(App)
    app = App()
    app = init_app(app, engine)
    morepath.run(app)



import time

import pytest
from pybrake.route_metric import RouteMetric
import pybrake.metrics as metrics
from pybrake.routes import _Routes, RouteStats, RouteStat, route_stat_key

metrics.FLUSH_PERIOD = 0
CONFIG = {
    "config": {
        "performance_stats": True,
        "error_host": "http://localhost:5000",
        "apm_host": "http://localhost:5000",
    }
}


def _test_setup(config: dict):
    routes = _Routes(**config)

    metric = RouteMetric(method="GET", route="/test")
    metric.status_code = 200
    metric.content_type = "application/json"
    metric.FLUSH_PERIOD = 5
    metric.end_time = time.time()

    key = route_stat_key(
        method=metric.method,
        route=metric.route,
        status_code=metric.status_code,
        time=metric.start_time,
    )

    stat = RouteStat(
        method=metric.method,
        route=metric.route,
        status_code=metric.status_code,
        time=metric.start_time,
    )
    routes.stats._stats = {key: stat}
    stat.add((metric.end_time - metric.start_time) * 1000)
    return routes


def test_routes_performance_stats():
    """
    To see what happens if performance_stats is set to false, run this test.
    :return:
    """
    routes = _Routes(**{"config": {
        "performance_stats": False,
    }})

    metric = RouteMetric(method="GET", route="/test")
    metric.status_code = 200
    metric.content_type = "application/json"
    metric.end_time = time.time()

    assert routes.notify(metric) is None


def test_routes_notify(mocker):
    mocker.patch(
        "pybrake.routes.RouteStats._flush", return_value=None
    )
    mocker.patch(
        "pybrake.routes.RouteBreakdowns._flush", return_value=None
    )
    routes = _Routes(**CONFIG)

    metric = RouteMetric(method="GET", route="/test")
    metric.status_code = 200
    metric.content_type = "application/json"
    metric.FLUSH_PERIOD = 0
    metric.end_time = time.time()
    assert routes.notify(metric) == None


def test_routes_flash_empty_stats():
    routes = _Routes(**CONFIG)

    metric = RouteMetric(method="GET", route="/test")
    metric.status_code = 200
    metric.content_type = "application/json"
    metric.end_time = time.time()
    with pytest.raises(ValueError, match=r"stats is empty"):
        routes.stats._flush()


def test_routes_flash_with_500():
    routes = _Routes(**CONFIG)

    metric = RouteMetric(method="GET", route="/test")
    metric.status_code = 500
    metric.content_type = "application/json"
    metric.end_time = time.time()
    with pytest.raises(ValueError, match=r"stats is empty"):
        routes.stats._flush()


def test_routes_flash_with_400():
    routes = _Routes(**CONFIG)

    metric = RouteMetric(method="GET", route="/test")
    metric.status_code = 400
    metric.content_type = "application/json"
    metric.end_time = time.time()
    with pytest.raises(ValueError, match=r"stats is empty"):
        routes.stats._flush()


def test_routes_flash_ok(mocker):
    mocker.patch(
        "pybrake.metrics.send",
        return_value=None
    )
    routes = _test_setup(CONFIG)
    assert routes.stats._flush() is None


def test_route_stat():
    metric = RouteMetric(method="GET", route="/test")
    metric.status_code = 200
    metric.content_type = "application/json"
    metric.end_time = time.time()

    stat = RouteStat(
        method=metric.method,
        route=metric.route,
        status_code=metric.status_code,
        time=1648551580.0367732,
    )

    assert stat.__dict__ == {
        'count': 0,
        'sum': 0,
        'sumsq': 0,
        'tdigest': 'AAAAAkAkAAAAAAAAAAAAAA==',
        'method': 'GET',
        'route': '/test',
        'statusCode': 200,
        'time': '2022-03-29T10:59:00Z'
    }


def test_routes_ab_url():
    CONFIG.update({'project_id': 0})
    stats = RouteStats(**CONFIG)
    assert stats._ab_url() == \
           'http://localhost:5000/api/v5/projects/0/routes-stats'



import os
from datetime import datetime

import cherrypy
import requests
from cp_sqlalchemy import SQLAlchemyTool, SQLAlchemyPlugin
from jinja2 import Environment, FileSystemLoader
from pybrake.middleware.cherrypy import PybrakePlugin
from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

config = {
    "PYBRAKE": {
        'project_id': 999999,
        'project_key': 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx',
        'environment': 'test'
    },
    '/': {
        'tools.db.on': True
    }
}

Base = declarative_base()

city_list = ["pune", "austin", "santabarbara", "washington"]

# GET CURRENT DIRECTORY
CUR_DIR = os.path.dirname(os.path.abspath(__file__))
env = Environment(loader=FileSystemLoader(CUR_DIR), trim_blocks=True)


class User(Base):
    __tablename__ = 'user'
    id = Column(Integer, primary_key=True)
    username = Column(String(80), unique=True, nullable=False)
    email = Column(String(120), unique=True, nullable=False)

    def __repr__(self):
        return "<User %r>" % self.username


class WeatherDetails(object):

    @property
    def db(self):
        return cherrypy.request.db

    # API for Hello Application
    @cherrypy.expose
    def index(self):
        res = self.db.query(User).all()
        if not res:
            self.db.add(User(username="test", email="test@test.com"))
            self.db.commit()
        template = env.get_template("index.html")
        return template.render()

    # API for current server date
    @cherrypy.expose
    @cherrypy.tools.json_out()
    def date(self):
        return {
            'date': "Current Date and Time is: %s" % datetime.now()
        }

    # API for location details
    @cherrypy.expose
    @cherrypy.tools.json_out()
    def locations(self):
        return {
            'locations': city_list
        }

    # API for weather details for a location
    @cherrypy.expose
    @cherrypy.tools.json_out()
    def weather(self, location_name):
        if location_name not in city_list:
            raise cherrypy.HTTPError(404, 'Location not found!')
        with requests.get(
                'https://airbrake.github.io/weatherapi/weather/' + location_name) as f:
            return f.json()


if __name__ == '__main__':
    cherrypy.tools.db = SQLAlchemyTool()
    cherrypy.tree.mount(WeatherDetails(), '/', config)

    sqlalchemy_plugin = SQLAlchemyPlugin(
        cherrypy.engine, Base, 'sqlite:///./sql_app.db'
    )
    sqlalchemy_plugin.subscribe()
    sqlalchemy_plugin.create()

    pybrake_plugin = PybrakePlugin(
        cherrypy.engine, **config.get('PYBRAKE')
    )
    pybrake_plugin.subscribe()
    pybrake_plugin.create()

    cherrypy.quickstart(WeatherDetails(), config=config)



import logging
import urllib


def get_exception():
    try:
        raise ValueError("hello")
    except ValueError as err:
        return err


def build_notice_from_str(notifier, s):
    return notifier.build_notice(s)


def build_logging_record_error(*args, **kwargs):
    logger, dh = logger_dummy_handler()
    logger.error(*args, **kwargs)
    return dh.record


def build_logging_record_exception():
    logger, dh = logger_dummy_handler()
    try:
        raise ValueError("hello")
    except ValueError as err:
        logger.exception(err)
        return dh.record


def logger_dummy_handler():
    class DummyHandler(logging.Handler):
        def __init__(self):
            super().__init__()
            self.record = None

        def emit(self, record):
            self.record = record

    logger = logging.getLogger("test")
    dh = DummyHandler()
    logger.addHandler(dh)

    return logger, dh


def get_nested_exception():
    try:
        raise get_exception()
    except ValueError as err:
        try:
            raise ValueError("world") from err
        except ValueError as subErr:
            return subErr


def get_exception_in_cython():
    # accumulation_tree is required by tdigest
    from accumulation_tree import AccumulationTree  # pylint: disable=import-outside-toplevel
    t = AccumulationTree(lambda x: x)
    t.insert(1, '1')
    try:
        return t.insert('1', '1')
    except TypeError as err:
        return err


class MockResponse(object):

    def __init__(self, resp_data, code=200, headers=None, msg='OK'):
        if headers is None:
            headers = {}
        self.resp_data = resp_data
        self.code = code
        self.msg = msg
        self.headers = {'content-type': 'text/plain; charset=utf-8'}
        self.headers.update(headers)

    def read(self):
        if self.resp_data == "IOError":
            raise IOError("IOError: for test")
        if self.resp_data == "HTTPError":
            raise urllib.error.HTTPError(*[None] * 5)
        return self.resp_data

    def getcode(self):
        return self.code


class TestBacklog(object):
    def append_stats(self, val, url, retry_count=0):
        pass



import time
from collections import deque
from threading import Timer

from . import metrics


class Backlog(Timer):
    def __init__(self, method, header, interval=60, maxlen=100,
                 error_notice=False, notifier=None, args=None, kwargs=None):
        super().__init__(interval, self.send, args, kwargs)
        self._backlog = deque(maxlen=maxlen)
        self._method = method
        self._header = header
        self._error_notice = error_notice
        self._notifier = notifier

    def send(self):
        # pop out queue items every 1 sec
        # (please ignore empty deque for now)
        while len(self._backlog) > 0:
            payload = self._backlog.popleft()
            if not self._error_notice:
                metrics.send(
                    url=payload.get('url'),
                    headers=self._header,
                    method=self._method,
                    payload=payload.get('data'),
                    retry_count=payload.get('retry_count') + 1
                )
            else:
                metrics.send_notice(
                    self._notifier,
                    notice=payload.get('data'),
                    url=payload.get('url'),
                    headers=self._header,
                    method=self._method,
                    retry_count=payload.get('retry_count') + 1
                )
            time.sleep(self.interval)

    def append_stats(self, val, url, retry_count=0):
        self._backlog.append({
            'retry_count': retry_count,
            'url': url,
            'data': val
        })
        if not self._started.is_set():
            self.start()



import time

import pytest
import pybrake.metrics as metrics
from pybrake.queries import QueryStat, QueryStats, query_stat_key

metrics.FLUSH_PERIOD = 0
CONFIG = {"config": {
    "performance_stats": True,
    "query_stats": True,
    "error_host": "http://localhost:5000",
    "apm_host": "http://localhost:5000",
}}
start_time = time.time()
query = "SELECT * FROM foos"
method = "GET"
route = '/test'


def _test_setup(config: dict):
    stats = QueryStats(**config)
    key = query_stat_key(
        query=query,
        method=method,
        route=route,
        time=start_time,
    )
    stat = QueryStat(
        query=query, method=method, route=route, time=start_time
    )
    stats._stats = {key: stat}
    return stats


def test_query_metric_start_end():
    stats = QueryStat(time=1648551580.0367732)
    assert stats.__dict__ == {
        'count': 0,
        'file': '',
        'function': '',
        'line': 0,
        'sum': 0,
        'sumsq': 0,
        'tdigest': 'AAAAAkAkAAAAAAAAAAAAAA==',
        'query': '',
        'method': '',
        'route': '',
        'time': '2022-03-29T10:59:00Z'
    }


def test_query_stats_performance_stats(mocker):
    """
    To see what happens if performance_stats is set to false, run this test.
    :return:
    """
    mocker.patch(
        "pybrake.queries.QueryStats._flush",
        return_value=None
    )
    stats = QueryStats(**{"config": {
        "performance_stats": False,
    }})
    assert stats.notify(
        query="SELECT * FROM foos",
        method="GET",
        route='test',
        start_time=time.time(),
        end_time=time.time(),
    ) is None


def test_query_stats_query_stats(mocker):
    """
    To see what happens if query_stats is set to false, run this test.
    :return:
    """
    mocker.patch(
        "pybrake.queries.QueryStats._flush",
        return_value=None
    )
    stats = QueryStats(**{"config": {
        "performance_stats": True,
        "query_stats": False,
    }})
    assert stats.notify(
        query="SELECT * FROM foos",
        method="GET",
        route='test',
        start_time=time.time(),
        end_time=time.time(),
    ) is None


def test_query_stats_notify_flash(mocker):
    mocker.patch(
        "pybrake.metrics.send",
        return_value=None
    )
    stats = _test_setup(CONFIG)
    assert stats._flush() is None
    # Empty Stats
    stats._stats = None
    with pytest.raises(ValueError, match=r"stats is empty"):
        stats._flush()


def test_query_stats_notify(mocker):
    mocker.patch(
        "pybrake.queries.QueryStats._flush",
        return_value=None
    )
    stats = QueryStats(**CONFIG)
    assert stats.notify(
        query="SELECT * FROM foos",
        method="GET",
        route='test',
        start_time=time.time(),
        end_time=time.time(),
    ) is None


def test_queue_ab_url():
    CONFIG.update({"project_id": 0})
    stats = QueryStats(**CONFIG)
    res = stats._ab_url()
    assert res == "http://localhost:5000/api/v5/projects/0/queries-stats"



import time
import traceback
from urllib.parse import quote as urllib_quote

from tg import hooks
from tg.appwrappers.errorpage import ErrorPageApplicationWrapper
from tg.controllers.decoratedcontroller import DecoratedController
from tg.request_local import request, response
from tg.configurator.components.error_reporting import \
    ErrorReportingConfigurationComponent
from tg import MinimalApplicationConfigurator

from ..metrics import (
    set_active as set_active_metrics,
    get_active as get_active_metrics,
    start_span,
    end_span,
)
from ..notifier import Notifier
from ..route_metric import RouteMetric

try:
    import backlash  # pylint: disable=unused-import
except ImportError:
    # pylint: disable=raise-missing-from
    raise ModuleNotFoundError('No module named "backlash". Backlash is used '
                              'to track error.')

try:
    # pylint: disable=ungrouped-imports
    from tg.configurator.components.sqlalchemy import \
        SQLAlchemyConfigurationComponent
    from sqlalchemy import event
except ImportError:
    _sqla_available = False
else:
    old_setup_sqlalchemy = SQLAlchemyConfigurationComponent.setup_sqlalchemy


    def _patch_setup_sqlalchemy(self, conf, app):
        old_setup_sqlalchemy(self, conf, app)
        _sqla_instrument(conf['tg.app_globals'].sa_engine, conf['pybrake'])


    SQLAlchemyConfigurationComponent.setup_sqlalchemy = _patch_setup_sqlalchemy

_UNKNOWN_ROUTE = "UNKNOWN"


class AirbrakeMiddleware:
    def __init__(self, notifier):
        self.notifier = notifier

    def report(self, traceback_ex):
        environ = traceback_ex.context.get('environ', {})
        is_backlash_event = getattr(traceback_ex.exc_value, 'backlash_event',
                                    False)
        if is_backlash_event:
            # Just a Stack Dump request from backlash
            _handle_exception(self.notifier, traceback_ex.exception, environ)
        else:
            # This is a real crash
            _handle_exception(self.notifier, traceback_ex.exception, environ)


def init_app(config):
    if 'pybrake' in config.configure():
        raise ValueError("pybrake is already injected")
    if 'PYBRAKE' not in config.configure():
        raise ValueError("app.settings['PYBRAKE'] is not defined")

    notifier = Notifier(**config.get_blueprint_value('PYBRAKE'))
    pybrake_config = notifier.config

    config.update_blueprint({
        'pybrake': notifier,
        # Exception handling
        "trace_errors.reporters": [AirbrakeMiddleware(notifier)],
    })
    # Exception handling
    if config.__class__ == MinimalApplicationConfigurator:
        config.register(ErrorReportingConfigurationComponent)
        config.register_application_wrapper(ErrorPageApplicationWrapper,
                                            after=True)

    # Route stats

    old_call = DecoratedController._call

    def _patch_call(self, action, params, remainder=None, context=None):
        if pybrake_config.get("performance_stats"):
            _before_request(request)
        try:
            return old_call(self, action, params, remainder, context)
        except Exception as er:
            context.response.status_code = 500
            raise er
        finally:
            if pybrake_config.get("performance_stats"):
                _after_request(notifier)
        # return res

    DecoratedController._call = _patch_call

    # Route Breakdown stats
    def before_template_render(*args, **kwargs):
        start_span("template")

    def after_template_render(*args, **kwargs):
        end_span("template")

    hooks.register('before_render_call', before_template_render)
    hooks.register('after_render_call', after_template_render)

    return config


def _sqla_instrument(engine, notifier):
    event.listen(engine, "before_cursor_execute", _before_cursor(notifier))
    event.listen(engine, "after_cursor_execute", _after_cursor(notifier))


def _before_cursor(notifier):
    def _sqla_before_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if not notifier.config.get("performance_stats"):
            return
        start_span("sql")

    return _sqla_before_cursor_execute


def _after_cursor(notifier):
    def _sqla_after_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if not notifier.config.get("performance_stats"):
            return
        end_span("sql")
        metric = get_active_metrics()
        if metric is not None:
            try:
                traceback_frm = traceback.extract_stack(limit=12)[0]
            except IndexError as er:  # pylint: disable=unused-variable
                traceback_frm = None
            notifier.queries.notify(
                query=statement,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=metric.start_time,
                end_time=time.time(),
                function=traceback_frm.name if traceback_frm else '',
                file=traceback_frm.filename if traceback_frm else '',
                line=traceback_frm.lineno if traceback_frm else 0,
            )

    return _sqla_after_cursor_execute


def _before_request(*remainder, **params):
    if request.controller_url:
        route = request.controller_url
    elif request.path:
        route = request.path
    else:
        route = _UNKNOWN_ROUTE
    metric = RouteMetric(method=request.method, route=route)
    set_active_metrics(metric)


def _after_request(notifier):
    metric = get_active_metrics()
    if metric is not None:
        metric.status_code = response.status_code
        metric.content_type = response.content_type
        metric.end_time = time.time()
        notifier.routes.notify(metric)
        set_active_metrics(None)


def get_host(environ):
    scheme = environ.get('wsgi.url_scheme')
    if 'HTTP_X_FORWARDED_HOST' in environ:
        result = environ['HTTP_X_FORWARDED_HOST']
    elif 'HTTP_HOST' in environ:
        result = environ['HTTP_HOST']
    else:
        result = environ['SERVER_NAME']
        if (scheme, str(environ['SERVER_PORT'])) not \
                in (('https', '443'), ('http', '80')):
            result += ':' + environ['SERVER_PORT']
    if result.endswith(':80') and scheme == 'http':
        result = result[:-3]
    elif result.endswith(':443') and scheme == 'https':
        result = result[:-4]
    return result


def get_current_url(environ, root_only=False, strip_querystring=False,
                    host_only=False):
    tmp = [environ['wsgi.url_scheme'], '://', get_host(environ)]
    cat = tmp.append
    if host_only:
        return ''.join(tmp) + '/'
    cat(urllib_quote(environ.get('SCRIPT_NAME', '').rstrip('/')))
    if root_only:
        cat('/')
    else:
        cat(urllib_quote('/' + environ.get('PATH_INFO', '').lstrip('/')))
        if not strip_querystring:
            qs = environ.get('QUERY_STRING')
            if qs:
                cat('?' + qs)
    return ''.join(tmp)


def request_filter(notice, environ):
    if request is None:
        return notice

    ctx = notice["context"]
    ctx["method"] = environ.get('REQUEST_METHOD')
    ctx["url"] = get_current_url(environ, strip_querystring=True)
    ctx["route"] = request.controller_url
    ctx["userAddr"] = request.client_addr

    if getattr(request, 'identity', None):
        user = {}
        for s in ["username", "name", "user_id"]:
            if hasattr(getattr(request, 'identity'), s):
                user[s] = getattr(getattr(request, 'identity'), s)
        ctx["user"] = user

    notice["params"]["request"] = dict(
        body=dict(request.body),
        data=environ.get(' wsgi.input'),
        params=dict(request.params),
        cookies=dict(request.cookies),
        headers=dict(request.headers),
        environ=environ,
        query_string=environ.get('QUERY_STRING'),
    )

    return notice


def _handle_exception(notifier, exception, environ):
    notice = notifier.build_notice(exception)
    notice = request_filter(notice, environ)
    notifier.send_notice(notice)



from datetime import datetime

import requests
from django.http import HttpResponse, JsonResponse
from rest_framework import status

# Create your views here.
city_list = ["pune", "austin", "santabarbara", "washington"]


# API for Hello Application
def hello(request):
    return HttpResponse("Hello, Welcome to the Weather App!")


# API for current server date
def getdate(request):
    return JsonResponse(
        data={
            "date": "Current Date and Time is: %s" % datetime.now()
        },
        status=status.HTTP_200_OK
    )


# API for location details
def get_location_details(request):
    return JsonResponse(
        data={
            "locations": city_list
        },
        status=status.HTTP_200_OK
    )


# API for weather details for a location
def get_weather_details(request, location_name):
    if location_name not in city_list:
        return JsonResponse(
            status=status.HTTP_404_NOT_FOUND,
            data={
                "error": "Location not found!"
            }
        )
    with requests.get("https://airbrake.github.io/weatherapi/weather/" +
                      location_name) as f:
        data = f.json()
    return JsonResponse(data, status=status.HTTP_200_OK, safe=False)



# -*- coding: utf-8 -*-
"""Setup the weather-fullstack application"""

import logging

from weather_fullstack.config.app_cfg import base_config

__all__ = ['setup_app']

log = logging.getLogger(__name__)

from .schema import setup_schema
from .bootstrap import bootstrap


def setup_app(command, conf, vars):
    """Place any commands to setup weather_fullstack here"""
    conf = base_config.configure(conf.global_conf, conf.local_conf)
    base_config.setup(conf)

    setup_schema(command, conf, vars)
    bootstrap(command, conf, vars)



from masonite.routes import Route

ROUTES = [
    Route.get("/", "WeatherController@show"),
    Route.get("/date", "WeatherController@date"),
    Route.get("/locations", "WeatherController@locations"),
    Route.get("/weather/@location_name:string", "WeatherController@weather"),
]



"""A WeatherController Module."""
from datetime import datetime

import requests
from app.models.User import User
from masonite.controllers import Controller
from masonite.response import Response
from masonite.views import View

city_list = ["pune", "austin", "santabarbara", "washington"]


class WeatherController(Controller):
    """WelcomeController Controller Class."""

    def show(self, view: View):
        users = User.all()
        if not users.count():
            user = User()
            users = user.register(dictionary={
                'name': 'Test',
                'email': 'test@test.com',
                'password': 'password',
            })
        print(users)
        return view.render("welcome")

    def date(self):
        return {
            'date': "Current Date and Time is: %s" % datetime.now()
        }

    def locations(self):
        return {
            'locations': city_list
        }

    def weather(self, location_name, response: Response):
        if location_name not in city_list:
            response.status(400)
            return {
                'error': 'Location not found!'
            }
        with requests.get(
                'https://airbrake.github.io/weatherapi/weather/' + location_name) as f:
            return f.json()



from masonite.foundation import Application, Kernel
from masonite.utils.location import base_path
from masonite.configuration import config

from Kernel import Kernel as ApplicationKernel

"""Start The Application Instance."""
application = Application(base_path())

"""Now Bind important providers needed to make the framework work."""
application.register_providers(Kernel, ApplicationKernel)

"""Now Bind important application specific providers needed to make the application work."""
application.add_providers(*config("providers.providers"))



import base64
import json
import threading

from . import constant
from . import metrics
from .backlog import Backlog
from .tdigest import as_bytes, TDigestStatGroups
from .utils import time_trunc_minute


class _RouteBreakdown(TDigestStatGroups):

    def __new__(cls, *, method="", route="", responseType="", time=None):
        instance = super(_RouteBreakdown, cls).__new__(cls)
        instance.__slots__ = instance.__slots__ + (
            "method", "route", "responseType", "time"
        )
        return instance

    def __init__(self, *, method="", route="", responseType="", time=None):
        super().__init__()
        self.method = method
        self.route = route
        self.responseType = responseType
        self.time = time_trunc_minute(time)

    @property
    def __dict__(self):
        b = as_bytes(self.td)
        self.tdigest = base64.b64encode(b).decode("ascii")
        d = {s: getattr(self, s) for s in self.__slots__ if s != "td"}

        groups = d["groups"]
        for k, v in groups.items():
            groups[k] = v.__dict__

        return d


class RouteBreakdowns:
    """
    RouteBreakdowns will collect request execution with breakdown statistics
    such as execution statistics of template rendering time and sql/nosql
    query.
    """

    def __init__(self, *, project_id=0, project_key="", **kwargs):
        self._config = kwargs["config"]

        self._project_id = project_id
        self._ab_headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + project_key,
        }
        self._env = kwargs.get("environment")

        self._thread = None
        self._lock = threading.Lock()
        self._stats = None
        self._backlog = None
        if self._config.get('backlog_enabled'):
            if not metrics.APM_Backlog:
                metrics.APM_Backlog = Backlog(
                    header=self._ab_headers,
                    method="POST",
                    maxlen=self._config.get('max_backlog_size'),
                )
            self._backlog = metrics.APM_Backlog

    def notify(self, metric):
        if not self._config.get("performance_stats"):
            return

        if (
            metric.status_code < 200
            or (metric.status_code >= 300 and metric.status_code < 400)
            or metric.status_code == 404
            or len(metric._groups) <= 1
        ):
            return

        if self._stats is None:
            self._stats = {}
            self._thread = threading.Timer(constant.FLUSH_PERIOD, self._flush)
            self._thread.start()

        key = metric._key()
        with self._lock:
            if key in self._stats:
                stat = self._stats[key]
            else:
                stat = _RouteBreakdown(
                    method=metric.method,
                    route=metric.route,
                    responseType=metric.response_type,
                    time=metric.start_time,
                )
                self._stats[key] = stat

            total_ms = (metric.end_time - metric.start_time) * 1000
            stat.add_groups(total_ms, metric._groups)

    def _flush(self):
        """
        TODO: Below disabled pylint will remove in refactoring
        """
        # pylint: disable=too-many-branches
        stats = None
        with self._lock:
            stats = self._stats
            self._stats = None

        if not stats:
            raise ValueError("stats is empty")

        out = {"routes": [v.__dict__ for v in stats.values()]}
        if self._env:
            out["environment"] = self._env

        out_json = json.dumps(out).encode("utf8")
        metrics.send(
            url=self._ab_url(), payload=out_json,
            headers=self._ab_headers, method="POST",
        )

    def _ab_url(self):
        return f"{self._config.get('apm_host')}/api/v5/projects/" \
               f"{self._project_id}/routes-breakdowns"


class RouteMetric(metrics.Metric):
    def __init__(self, *, method="", route="", status_code=0, content_type=""):
        super().__init__()
        self.method = method
        self.route = route
        self.status_code = status_code
        self.content_type = content_type
        self.start_span(constant.HTTP_HANDLER, start_time=self.start_time)

    def end(self):
        super().end()
        self.end_span(constant.HTTP_HANDLER, end_time=self.end_time)

    def _key(self):
        time = self.start_time // 60 * 60
        return (self.method, self.route, self.response_type, time)

    @property
    def response_type(self):
        if self.status_code >= 500:
            return "5xx"
        if self.status_code >= 400:
            return "4xx"
        return self.content_type.split(";")[0].split("/")[-1]



# -*- coding: utf-8 -*-
"""Main Controller"""

from datetime import datetime

import requests
from tg import expose
from tg import tmpl_context
from tg.exceptions import HTTPNotFound
from weather_fullstack import model
from weather_fullstack.controllers.error import ErrorController
from weather_fullstack.lib.base import BaseController
from weather_fullstack.model import DBSession

__all__ = ['RootController']

city_list = ["pune", "austin", "santabarbara", "washington"]


class RootController(BaseController):
    """
    The root controller for the weather-fullstack application.

    All the other controllers and WSGI applications should be mounted on this
    controller. For example::

        panel = ControlPanelController()
        another_app = AnotherWSGIApplication()

    Keep in mind that WSGI applications shouldn't be mounted directly: They
    must be wrapped around with :class:`tg.controllers.WSGIAppController`.

    """
    error = ErrorController()

    def _before(self, *args, **kw):
        tmpl_context.project_name = "weather_fullstack"

    @expose("weather_fullstack.templates.index")
    def index(self):
        count = DBSession.query(model.auth.User).count()
        if not count:
            user = model.auth.User()
            user.user_name = "test"
            user.email_address = "test@test.com"
            user.display_name = "test"
            user.password = "test"
            DBSession.add(user)
        return dict(app="Weather")

    @expose("json")
    def date(self):
        return dict(date="Current Date and Time is: %s" % datetime.now())

    @expose("json")
    def locations(self):
        return dict(locations=city_list)

    @expose("json")
    def weather(self, location_name):
        if location_name not in city_list:
            raise HTTPNotFound(detail="Location not found!")
        with requests.get(
                'https://airbrake.github.io/weatherapi/weather/' + location_name) as f:
            return f.json()



import base64
import json
import threading

from . import metrics
from . import constant
from .backlog import Backlog
from .tdigest import TDigestStat, as_bytes
from .utils import time_trunc_minute


class QueryStat(TDigestStat):

    def __new__(cls, *, query="", method="", route="", function="",
                 file="", line=0, time):
        instance = super(QueryStat, cls).__new__(cls)
        instance.__slots__ = instance.__slots__ + (
            "query", "method", "route", "time", "function", "file", "line"
        )
        return instance

    @property
    def __dict__(self):
        tdigest = as_bytes(self.td)
        self.tdigest = base64.b64encode(tdigest).decode("ascii")

        return {s: getattr(self, s) for s in self.__slots__ if s != "td"}

    def __init__(self, *, query="", method="", route="", function="",
                 file="", line=0, time=None):
        super().__init__()
        self.query = query
        self.method = method
        self.route = route
        self.function = function
        self.file = file
        self.line = line
        self.time = time_trunc_minute(time)


class QueryStats:
    """
    Query stats are transmitted to the Airbrake site using QueryStats.
    QueryStat will collect query execution statistics such as query start time,
     end time, query statement, query route, and route method.
    """

    def __init__(self, *, project_id=0, project_key="", **kwargs):
        self._config = kwargs["config"]

        self._project_id = project_id
        self._ab_headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + project_key,
        }
        self._env = kwargs.get("environment")

        self._thread = None
        self._lock = threading.Lock()
        self._stats = None
        self._backlog = None
        if self._config.get('backlog_enabled'):
            if not metrics.APM_Backlog:
                metrics.APM_Backlog = Backlog(
                    header=self._ab_headers,
                    method="POST",
                    maxlen=self._config.get('max_backlog_size'),
                )
            self._backlog = metrics.APM_Backlog

    def notify(
            self, *, query="", method="", route="", start_time=None,
            function="",  file="", line=0, end_time=None
    ):
        if not self._config.get("performance_stats"):
            return
        if not self._config.get("query_stats"):
            return

        key = query_stat_key(
            query=query, method=method, route=route, time=start_time,
            function=function, file=file, line=line
        )
        ms = (end_time - start_time) * 1000

        with self._lock:
            if self._stats is None:
                self._stats = {}
                self._thread = threading.Timer(
                    constant.FLUSH_PERIOD, self._flush)
                self._thread.start()

            if key in self._stats:
                stat = self._stats[key]
            else:
                stat = QueryStat(
                    query=query, method=method, route=route, time=start_time,
                    function=function, file=file, line=line
                )
                self._stats[key] = stat
            stat.add(ms)

    def _flush(self):
        """
        TODO: Below disabled pylint will remove in refactoring
        """
        # pylint: disable=too-many-branches
        stats = None
        with self._lock:
            stats = self._stats
            self._stats = None

        if not stats:
            raise ValueError("stats is empty")

        out = {"queries": [v.__dict__ for v in stats.values()]}
        if self._env:
            out["environment"] = self._env

        out = json.dumps(out).encode("utf8")
        metrics.send(
            url=self._ab_url(), payload=out,
            headers=self._ab_headers, method="POST",
        )

    def _ab_url(self):
        return f"{self._config.get('apm_host')}/api/v5/projects" \
               f"/{self._project_id}/queries-stats"


def query_stat_key(*, query="", method="", route="", function="", file="",
                   line=0, time=None):
    time = time // 60 * 60
    return query, method, route, function, file, line, time



import datetime
import json
import os
from datetime import datetime

import hug
import requests
from falcon import HTTP_400, HTTP_500, HTTPError
from pybrake.middleware.hug import init_app

from sqlalchemy import INTEGER, Column, String, select, insert
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base

engine = create_engine('sqlite:///./test.db')
Base = declarative_base()

PYBRAKE_CONFIG = {
    "PYBRAKE": {
        "project_id": 99999,
        "project_key": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    }
}

api = hug.API(__name__)
PYBRAKE_CONFIG = init_app(api, PYBRAKE_CONFIG, engine)

DIRECTORY = os.path.dirname(os.path.realpath(__file__))
city_list = ["pune", "austin", "santabarbara", "washington"]


class User(Base):
    __tablename__ = "user"

    id = Column(INTEGER, primary_key=True)
    username = Column(String(80), unique=True, nullable=False)
    email = Column(String(120), unique=True, nullable=False)

    def __repr__(self):
        return "<User %r>" % self.username


@hug.get("/", output=hug.output_format.html, api=api)
def index(**kwargs):
    with engine.begin() as conn:
        stmt = select(User)
        result = conn.execute(stmt)
        res = result.scalar()
        if not res:
            stmt = insert(User).values(username="test",
                                       email="test@test.com")
            conn.execute(stmt)
    with open(os.path.join(DIRECTORY, "templates/index.html")) as document:
        return document.read()


@hug.get("/date", api=api)
def date():
    return {"date": datetime.now()}


@hug.get("/locations", api=api)
def locations():
    return {"locations": city_list}


@hug.get("/weather/{location_name}", api=api)
def get_weather_details(location_name, resp):
    if location_name not in city_list:
        resp.status = HTTP_400
        resp.text = json.dumps({'error': 'Location not found!'})
        return
    if location_name == "washington":
        resp.status = HTTP_500
        resp.text = json.dumps(
            {'error': "Washington's data is not available!"})
        raise HTTPError(
            status=HTTP_500,
            title=str("Washington's data is not available!")
        )

    with requests.get(
            'https://airbrake.github.io/weatherapi/weather/' + location_name) as f:
        resp.status = f.status_code
        resp.text = f.content


if __name__ == "__main__":
    Base.metadata.create_all(engine)
    hug.API(__name__).http.serve()



from warnings import warn, simplefilter

from .middleware import celery


def __getattr__(name):
    if name == 'patch_celery':
        simplefilter("always", DeprecationWarning)
        warn(DeprecationWarning('pybrake.celery has been '
                                'renamed to '
             'pybrake.middleware.celery'))
        return celery.patch_celery
    raise AttributeError('No module named ' + name)



import base64
import json
import threading

from . import constant
from . import metrics
from .backlog import Backlog
from .tdigest import as_bytes, TDigestStatGroups
from .utils import time_trunc_minute


class QueueMetric(metrics.Metric):
    def __init__(self, *, queue=""):
        super().__init__()
        self.queue = queue
        self.start_span(constant.QUEUE_HANDLER, start_time=self.start_time)

    def end(self):
        super().end()
        self.end_span(constant.QUEUE_HANDLER, end_time=self.end_time)

    def _key(self):
        time = self.start_time // 60 * 60
        return (self.queue, time)


class _QueueStat(TDigestStatGroups):

    def __new__(cls, *, queue="", time=None):
        instance = super(_QueueStat, cls).__new__(cls)
        instance.__slots__ = instance.__slots__ + (
            "queue", "time"
        )
        return instance

    def __init__(self, *, queue="", time=None):
        super().__init__()
        self.queue = queue
        self.time = time_trunc_minute(time)

    @property
    def __dict__(self):
        b = as_bytes(self.td)
        self.tdigest = base64.b64encode(b).decode("ascii")
        d = {s: getattr(self, s) for s in self.__slots__ if s != "td"}

        groups = d["groups"]
        for k, v in groups.items():
            groups[k] = v.__dict__

        return d


class QueueStats:
    """
    Queue (background task, cron task) stats are transmitted to the Airbrake
    site using QueueStats.
    QueueStats will collect query execution statistics such as queue task
    start time, end time, task name.
    """

    def __init__(self, *, project_id=0, project_key="", **kwargs):
        self._config = kwargs["config"]

        self._project_id = project_id
        self._ab_headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + project_key,
        }
        self._env = kwargs.get("environment")

        self._thread = None
        self._lock = threading.Lock()
        self._stats = None
        self._backlog = None
        if self._config.get('backlog_enabled'):
            if not metrics.APM_Backlog:
                metrics.APM_Backlog = Backlog(
                    header=self._ab_headers,
                    method="POST",
                    maxlen=self._config.get('max_backlog_size'),
                )
            self._backlog = metrics.APM_Backlog

    def notify(self, metric):
        if not self._config.get("performance_stats"):
            return
        if not self._config.get("queue_stats"):
            return
        if len(metric._groups) <= 1:
            return

        metric.end()

        key = metric._key()
        with self._lock:
            if self._stats is None:
                self._stats = {}
                self._thread = threading.Timer(constant.FLUSH_PERIOD,
                                               self._flush)
                self._thread.start()

            if key in self._stats:
                stat = self._stats[key]
            else:
                stat = _QueueStat(queue=metric.queue, time=metric.start_time)
                self._stats[key] = stat

            total_ms = (metric.end_time - metric.start_time) * 1000
            stat.add_groups(total_ms, metric._groups)

    def _flush(self):
        """
        TODO: Below disabled pylint will remove in refactoring
        """
        # pylint: disable=too-many-branches
        stats = None
        with self._lock:
            stats = self._stats
            self._stats = None

        if not stats:
            raise ValueError("stats is empty")

        out = {"queues": [v.__dict__ for v in stats.values()]}
        if self._env:
            out["environment"] = self._env

        out_json = json.dumps(out).encode("utf8")
        metrics.send(
            url=self._ab_url(), headers=self._ab_headers,
            payload=out_json, method="POST",
        )

    def _ab_url(self):
        return f"{self._config.get('apm_host')}/api/v5/projects/{self._project_id}/queues-stats"



import json
import os
from datetime import datetime
from wsgiref.simple_server import make_server

import falcon
import falcon_sqla
import jinja2
import requests
from pybrake.middleware.falcon import init_app
from sqlalchemy import INTEGER, Column, String, select, insert
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base

engine = create_engine('sqlite:///./test.db')
Base = declarative_base()
manager = falcon_sqla.Manager(engine)

app = falcon.App(middleware=[manager.middleware])

PYBRAKE_CONFIG = {
    "PYBRAKE": {
        "project_id": 999999,
        "project_key": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    }
}

# _base_model_session_ctx = ContextVar("session")

app, PYBRAKE_CONFIG = init_app(app, PYBRAKE_CONFIG, engine)


class User(Base):
    __tablename__ = "user"

    id = Column(INTEGER, primary_key=True)
    username = Column(String(80), unique=True, nullable=False)
    email = Column(String(120), unique=True, nullable=False)

    def __repr__(self):
        return "<User %r>" % self.username


city_list = ["pune", "austin", "santabarbara", "washington"]


def load_template(name):
    path = os.path.join('templates', name)
    with open(os.path.abspath(path), 'r') as fp:
        return jinja2.Template(fp.read())


# API for Hello Application
class Weather:
    def on_get(self, req, resp):
        template = load_template('index.html')
        resp.status = falcon.HTTP_200  # This is the default status
        resp.content_type = falcon.MEDIA_HTML  # Default is JSON, so override

        with engine.begin() as conn:
            stmt = select(User)
            result = conn.execute(stmt)
            res = result.scalar()
            if not res:
                stmt = insert(User).values(username="test",
                                           email="test@test.com")
                conn.execute(stmt)

        resp.text = template.render()


# API for current server date
class Date:
    def on_get(self, req, resp):
        resp.status = falcon.HTTP_200
        resp.text = json.dumps({
            'date': "Current Date and Time is: %s" % datetime.now()
        })


# API for location details
class Locations:
    def on_get(self, req, resp):
        resp.status = falcon.HTTP_200
        resp.text = json.dumps({'locations': city_list})


# API for weather details for a location
class WeatherLocation:
    def on_get(self, req, resp, location_name):
        if location_name not in city_list:
            resp.status = falcon.HTTP_400
            resp.text = json.dumps({'error': 'Location not found!'})
            return
        if location_name == "washington":
            resp.status = falcon.HTTP_500
            resp.text = json.dumps({'error': "Washington's data is not available!"})
            raise falcon.HTTPError(
                status=falcon.HTTP_500,
                title=str("Washington's data is not available!")
            )

        with requests.get(
                'https://airbrake.github.io/weatherapi/weather/' + location_name) as f:
            resp.status = f.status_code
            resp.text = f.content


weather = Weather()
app.add_route('/', weather)

date = Date()
app.add_route('/date', date)

locations = Locations()
app.add_route('/locations', locations)

weather_location = WeatherLocation()
app.add_route('/weather/{location_name}', weather_location)

# debug logs enabled with debug = True
if __name__ == "__main__":
    with make_server('', 8000, app) as httpd:
        print('Serving on port 8000...')
        Base.metadata.create_all(engine)

        # Serve until process is killed
        httpd.serve_forever()



from datetime import datetime

import aiohttp_jinja2
import jinja2
import requests
from aiohttp import web
from pybrake.middleware.aiohttp import pybrake_middleware
from sqlalchemy import Column, Integer, String, select, insert
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base

city_list = ["austin", "pune", "santabarbara", "washington"]

routes = web.RouteTableDef()

# SQL setup
SQLALCHEMY_DATABASE_URL = "sqlite:///./sql_app.db"

engine = create_engine(
    SQLALCHEMY_DATABASE_URL, connect_args={"check_same_thread": False}
)

Base = declarative_base(bind=engine)

pybrake_middleware = pybrake_middleware(sqlEngine=engine)


class User(Base):
    __tablename__ = 'user'
    id = Column(Integer, primary_key=True)
    username = Column(String(80), unique=True, nullable=False)
    email = Column(String(120), unique=True, nullable=False)

    def __repr__(self):
        return "<User %r>" % self.username


@routes.get('/')
def index(request):
    with engine.begin() as conn:
        stmt = select(User)
        result = conn.execute(stmt)
        res = result.scalar()
        if not res:
            stmt = insert(User).values(username="test", email="test@test.com")
            conn.execute(stmt)
    return aiohttp_jinja2.render_template('index.html', request, {})


@routes.get('/date')
def getdate(request):
    return web.json_response(data={
        "date": "Current date and time is: %s" % datetime.now()
    })


@routes.get('/locations')
def get_location_details(request):
    return web.json_response(data={
        'cities': city_list
    })


# API for weather details for a location
@routes.get('/weather/{location_name}')
def get_weather_details(request):
    location_name = request.match_info['location_name']
    if location_name not in city_list:
        return web.json_response(data={
            'error': 'Not found: Location not found!'
        }, status=404)
    with requests.get(
            'https://airbrake.github.io/weatherapi/weather/' + location_name) as f:
        data = f.json()
        return web.json_response(data=data)


if __name__ == '__main__':
    app = web.Application(middlewares=[pybrake_middleware])
    aiohttp_jinja2.setup(app,
                         loader=jinja2.FileSystemLoader(
                             './templates'))
    app['PYBRAKE'] = dict(
        project_id=999999,
        project_key='xxxxxxxxxxxxxxxxxxxxx',
        environment='develop'  # optional
    )
    app.add_routes(routes)
    Base.metadata.create_all(engine)
    web.run_app(app)



import contextvars
import logging
import time
import traceback
import types

from starlette.applications import Starlette
from starlette.exceptions import ExceptionMiddleware
from starlette.responses import Response
from starlette.routing import Match
from starlette.types import Receive, Scope, Send

from ..metrics import (
    set_active as set_active_metrics,
    get_active as get_active_metrics,
    start_span,
    end_span,
)
from ..notifier import Notifier
from ..route_metric import RouteMetric

logger = logging.Logger(__name__)

current_request = contextvars.ContextVar("request_global",
                                         default=types.SimpleNamespace())

try:
    from sqlalchemy import event
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True

_UNKNOWN_ROUTE = "UNKNOWN"


async def form_data(req):
    return await req.form()


def request_filter(notice):
    request = current_request.get()
    request_values = dict(request.items())
    if request is None:
        return notice

    ctx = notice["context"]
    ctx["method"] = request.method
    ctx["url"] = request.url
    ctx["userAddr"] = request.client.host

    try:
        user = request.user
    except AssertionError:
        user = {}

    ctx["user"] = user

    try:
        session = request.session
    except AssertionError:
        session = {}

    notice["params"]["request"] = dict(
        session=session,
        cookies=dict(request.cookies),
        headers=dict(request.headers),
        path=request_values.get('path'),
        path_params=dict(request.path_params),
        query_params=dict(request.query_params),
    )

    return notice


def _before_request(request):
    request_route = None
    for route in request.app.router.routes:
        match, _ = route.matches(request)
        if match == Match.FULL:
            request_route = route.path_format

    if request_route:
        route = request_route
    elif request.url:
        route = request.url.path
    else:
        route = _UNKNOWN_ROUTE

    metric = RouteMetric(method=request.method, route=route)
    set_active_metrics(metric)


def _after_request(response, notifier):
    metric = get_active_metrics()
    if metric is not None:
        metric.status_code = response.status_code
        metric.content_type = response.headers.get("Content-Type")
        metric.end_time = time.time()
        notifier.routes.notify(metric)
        set_active_metrics(None)


def _before_cursor(notifier):
    def _sqla_before_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if notifier and not notifier.config.get("performance_stats"):
            return
        start_span("sql")

    return _sqla_before_cursor_execute


def _after_cursor(notifier):
    def _sqla_after_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if notifier and not notifier.config.get("performance_stats"):
            return
        end_span("sql")
        metric = get_active_metrics()
        if metric is not None:
            try:
                traceback_frm = traceback.extract_stack(limit=10)[0]
            except IndexError as er:  # pylint: disable=unused-variable
                traceback_frm = None
            notifier.queries.notify(
                query=statement,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=metric.start_time,
                end_time=time.time(),
                function=traceback_frm.name if traceback_frm else '',
                file=traceback_frm.filename if traceback_frm else '',
                line=traceback_frm.lineno if traceback_frm else 0,
            )

    return _sqla_after_cursor_execute


def init_app(app, sqlEngine=None) -> Starlette:
    """
    Initiate the pybrake notifier and apply the patch for
    error monitoring and APM.
    :param app: instance of Starlette application
    :param sqlEngine: SQLALCHEMY engine instance
    :return: Starlette application instance after apply patch or new setting
    """
    if getattr(app.state, "pybrake", None) is not None:
        raise ValueError("pybrake is already injected")
    if getattr(app.state, "PYBRAKE", None) is None:
        raise ValueError("app.state.PYBRAKE is not defined")

    app, notifier = init_pybrake(app, app.state.PYBRAKE, sqlEngine)
    app.state.pybrake = notifier
    return app


def init_pybrake(app, config, sqlEngine=None):
    @app.middleware("http")
    async def init_request_vars(request, call_next):
        current_request.set(request)
        response = await call_next(request)
        return response

    notifier = Notifier(**config)
    config = notifier.config

    notifier.add_filter(request_filter)

    # Error notification patch
    old_exception_error_call = ExceptionMiddleware.__call__

    async def patch_exception_error_call(
            self, scope: Scope, receive: Receive, send: Send) -> None:
        try:
            await old_exception_error_call(self, scope, receive, send)
        except Exception as exc:
            notifier.notify(exc)
            raise exc

    ExceptionMiddleware.__call__ = patch_exception_error_call

    # Route Stats monitoring
    @app.middleware("http")
    async def process_route_stats(request, call_next):
        if config.get("performance_stats"):
            _before_request(request)
        try:
            response = await call_next(request)
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(e)
            response = Response(content=traceback.format_exc(),
                                status_code=500)
        if config.get("performance_stats"):
            _after_request(response, notifier)
        return response

    # Route breakdown stats monitoring
    old_render = Response.render

    def patch_render(self, content) -> bytes:
        if self.media_type == 'text/html':
            start_span("template")
        res = old_render(
            self, content
        )
        if self.media_type == 'text/html':
            end_span("template")
        return res

    Response.render = patch_render

    # Query Stats monitoring
    if _sqla_available and sqlEngine is not None:
        event.listen(sqlEngine, "before_cursor_execute",
                     _before_cursor(notifier))
        event.listen(sqlEngine, "after_cursor_execute",
                     _after_cursor(notifier))

    return app, notifier



import logging
import traceback
import sys
import time

from .. import Notifier
from .. import RouteMetric
from ..metrics import (
    set_active as set_active_metrics,
    get_active as get_active_metrics,
    start_span,
    end_span,
)

logger = logging.getLogger(__name__)

try:
    from pyramid.httpexceptions import HTTPException
    from pyramid.request import Request
    from pyramid.threadlocal import get_current_request
    from pyramid.config import Configurator
    from pyramid import renderers, router
except ImportError as e:
    logger.error(str(e))
    sys.exit()

try:
    from sqlalchemy import event
    from pyramid_basemodel import Base
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True

try:
    from pyramid_fullauth.request import request_user as current_user
except ImportError:
    _pyramid_login_available = False
else:
    _pyramid_login_available = True

_UNKNOWN_ROUTE = "UNKNOWN"


def request_filter(notice):
    request = get_current_request()

    if request is None:
        return notice

    ctx = notice["context"]
    ctx["method"] = request.method
    ctx["url"] = request.url
    ctx["route"] = str(request.path)
    ctx["userAddr"] = request.remote_addr

    if _pyramid_login_available:
        curr_user = current_user(request)
        user = dict(id=curr_user.id)
        for s in ["username", "firstname", "lastname"]:
            if hasattr(curr_user, s):
                user[s] = getattr(curr_user, s)
        ctx["user"] = user

    notice["params"]["Request"] = dict(
        get=dict(request.GET),
        post=dict(request.POST),
        params=dict(request.params),
        body=dict(request.body),
        files=dict(request.body_file),
        cookies=dict(request.cookies),
        headers=dict(request.headers),
        environ=request.environ,
        view_args=request.urlargs,
    )

    return notice


def _sqla_instrument(notifier):
    try:
        sqla = Base.metadata.bind
    except Exception as err:  # pylint: disable=broad-except
        raise err
    event.listen(sqla.engine, "before_cursor_execute",
                 _before_cursor(notifier))
    event.listen(sqla.engine, "after_cursor_execute", _after_cursor(notifier))


def _before_cursor(notifier):
    def _sqla_before_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if notifier and not notifier.config.get("performance_stats"):
            return
        start_span("sql")

    return _sqla_before_cursor_execute


def _after_cursor(notifier):
    def _sqla_after_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if notifier and not notifier.config.get("performance_stats"):
            return
        end_span("sql")
        metric = get_active_metrics()
        if metric is not None:
            try:
                traceback_frm = traceback.extract_stack(limit=10)[0]
            except IndexError as er:  # pylint: disable=unused-variable
                traceback_frm = None
            notifier.queries.notify(
                query=statement,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=metric.start_time,
                end_time=time.time(),
                function=traceback_frm.name if traceback_frm else '',
                file=traceback_frm.filename if traceback_frm else '',
                line=traceback_frm.lineno if traceback_frm else 0,
            )

    return _sqla_after_cursor_execute


def route_stats_tween_factory(handler, registry):
    notifier = registry.settings.get('pybrake')
    if notifier.config.get('performance_stats'):
        # if stats support is enabled, return a wrapper
        def stats_tween(request):
            response = None
            metric = RouteMetric(
                method=request.method,
                route=str(request.path)
            )
            set_active_metrics(metric)
            try:
                response = handler(request)
            finally:
                metric = get_active_metrics()
                if metric is not None and response:
                    metric.status_code = response.status_code
                    metric.content_type = response.headers.get("Content-Type")
                    metric.end_time = time.time()
                    notifier.routes.notify(metric)
                    set_active_metrics(None)
            return response

        return stats_tween
    # if stats support is not enabled, return the original
    # handler
    return handler


def init_pybrake_config(config):
    if "pybrake" in config.registry.settings:
        raise ValueError("pybrake is already injected")
    if "PYBRAKE" not in config.registry.settings:
        raise ValueError("config.registry.settings['PYBRAKE'] is not defined")

    notifier = Notifier(**config.registry.settings["PYBRAKE"])

    notifier.add_filter(request_filter)
    config.registry.settings.update({"pybrake": notifier})

    # Error Notification Patch
    if hasattr(Request, "invoke_exception_view"):

        old_handle_exception_view = Request.invoke_exception_view
        Request.old_handle_exception_view = old_handle_exception_view

        def _handle_exception(self, *args, **kwargs):
            try:
                res = old_handle_exception_view(self, *args, **kwargs)
            except HTTPException as err:
                notifier.notify(err)
                raise err
            if res and res.status_int == 500 and self.exception:
                notifier.notify(res)
            return res

        Request.invoke_exception_view = _handle_exception

    # Patch for error notification on route call
    old_route_call = router.Router.__call__

    def _handle_exception_route_call(self, environ, start_response):
        try:
            return old_route_call(self, environ, start_response)
        except Exception as err:
            notifier.notify(err)
            raise err
    router.Router.__call__ = _handle_exception_route_call

    # Route Stats patch
    config.add_tween(
        'pybrake.middleware.pyramid.route_stats_tween_factory')

    # Patch for Template Render Stats
    old_render = renderers.render
    old_render_to_response = renderers.render_to_response

    def patch_render(*args, **kwargs):
        start_span("template")
        res = old_render(*args, **kwargs)
        end_span("template")
        return res

    def patch_render_to_response(*args, **kwargs):
        start_span("template")
        res = old_render_to_response(*args, **kwargs)
        end_span("template")
        return res

    renderers.render = patch_render
    renderers.render_to_response = patch_render_to_response

    # Patch for set up query stats
    old_make_wsgi_app = Configurator.make_wsgi_app

    def patch_make_wsgi_app(self):
        res = old_make_wsgi_app(self=self)
        notifier = res.registry.settings.get('pybrake')
        if _sqla_available and notifier:
            _sqla_instrument(notifier)

        return res

    Configurator.make_wsgi_app = patch_make_wsgi_app

    return config



from django.urls import path
from . import views

urlpatterns = [
    path('', views.hello),
    path('date', views.getdate),
    path('locations', views.get_location_details),
    path('weather/<location_name>', views.get_weather_details),
]



from warnings import warn, simplefilter

from .middleware import django


def __getattr__(name):
    if name == 'AirbrakeMiddleware':
        simplefilter("always", DeprecationWarning)
        warn(DeprecationWarning('pybrake.django.AirbrakeMiddleware has been '
                                'renamed to '
             'pybrake.middleware.django.AirbrakeMiddleware'))
        return django.AirbrakeMiddleware
    raise AttributeError('No module named ' + name)



"""Initial Schema

Revision ID: 462c8a652e3f
Revises: None
Create Date: 2022-09-22 20:49:04.054205

"""

# revision identifiers, used by Alembic.
revision = '462c8a652e3f'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'tg_user',
        sa.Column('user_id', sa.Integer, primary_key=True),
        sa.Column('user_name', sa.String(16), nullable=False),
        sa.Column('email_address', sa.Unicode(255)),
        sa.Column('display_name', sa.Unicode(255)),
        sa.Column('password', sa.Unicode(128)),
        sa.Column('created', sa.Date()),
    )


def downgrade():
    pass



import pytest

import pybrake.metrics as metrics
from pybrake.queues import QueueMetric, QueueStats, _QueueStat

metrics.FLUSH_PERIOD = 0


def test_queue_metric_start_end():
    metric = QueueMetric(queue="test")
    metric.end()
    assert metric.end_time is not None


def test_queue_metric_key():
    metric = QueueMetric(queue="test")
    assert ("test", metric.start_time // 60 * 60) == metric._key()


def test_queue_stats_performance_stats(mocker):
    mocker.patch(
        "pybrake.queues.QueueStats._flush",
        return_value=None
    )
    metric = QueueMetric(queue="foo_queue")
    metric._groups = {'redis': 24.0, 'sql': 0.4}
    stats = QueueStats(**{"config": {"performance_stats": False}})
    metric.end()
    assert stats.notify(metric) is None


def test_queue_stats_queue_stat_value():
    metric = QueueMetric(queue="foo_queue")
    stat = _QueueStat(queue=metric.queue, time=metric.start_time)
    assert stat.__dict__.get('queue') == 'foo_queue'


def test_queue_stats_queue_stats(mocker):
    """
    To see what happens if performance_stats is set to false, run this test.
    :return:
    """
    mocker.patch(
        "pybrake.queues.QueueStats._flush",
        return_value=None
    )
    metric = QueueMetric(queue="foo_queue")
    stats = QueueStats(**{"config": {
        "performance_stats": False,
    }})
    metric.end()
    assert stats.notify(metric) is None


def test_queue_stats_false(mocker):
    """
    To see what happens if queue_stats is set to false, run this test.
    :return:
    """
    mocker.patch(
        "pybrake.queues.QueueStats._flush",
        return_value=None
    )
    metric = QueueMetric(queue="foo_queue")
    stats = QueueStats(**{"config": {
        "performance_stats": True,
        "queue_stats": False,
    }})
    metric.end()
    assert stats.notify(metric) is None


def test_queue_stats_notify(mocker):
    mocker.patch(
        "pybrake.queues.QueueStats._flush",
        return_value=None
    )
    metric = QueueMetric(queue="foo_queue")
    metric._groups = {'redis': 24.0, 'sql': 0.4}
    stats = QueueStats(**{"config": {
        "performance_stats": True,
        "queue_stats": True,
        "error_host": "https://api.airbrake.io",
        "apm_host": "https://api.airbrake.io",
    }})
    metric.end()
    assert stats.notify(metric) is None


def test_queue_ab_url():
    stats = QueueStats(**{"config": {
        "performance_stats": True,
        "queue_stats": True,
        "error_host": "https://api.airbrake.io",
        "apm_host": "https://api.airbrake.io",
    }})
    assert stats._ab_url() == \
           'https://api.airbrake.io/api/v5/projects/0/queues-stats'


def test_queue_flush_blank_stet():
    stats = QueueStats(**{"config": {
        "performance_stats": True,
        "queue_stats": True
    }})
    with pytest.raises(ValueError, match=r"stats is empty"):
        stats._flush()



import time
from typing import (
    Dict,
    Optional,
)

from sanic import Sanic
from sanic import response
from sanic.handlers import ErrorHandler

from ..metrics import (
    set_active as set_active_metric,
    get_active as get_active_metric,
    start_span, end_span)
from ..notifier import Notifier
from ..route_metric import RouteMetric

try:
    from sqlalchemy import event
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True

_UNKNOWN_ROUTE = "UNKNOWN"


def request_filter(request, notice):
    if request is None:
        return notice
    ctx = notice["context"]
    ctx["method"] = request.method
    ctx["url"] = request.url
    ctx["route"] = str(request.route)
    ctx["userAddr"] = request.remote_addr if request.remote_addr else \
        request.ip

    user = {}
    try:
        if request.parsed_credentials:
            for s in ["username", "name"]:
                if hasattr(request.parsed_credentials, s):
                    user[s] = getattr(request.parsed_credentials, s)
            ctx["user"] = user
    except AttributeError:
        ctx["user"] = user

    notice["params"]["request"] = dict(
        query_args=request.query_args,
        query_string=request.query_string,
        form=dict(request.form),
        forwarded=dict(request.forwarded),
        json=dict(request.json) if request.json else {},
        files=dict(request.files),
        cookies=dict(request.cookies),
        headers=dict(request.headers),
        endpoint=request.name,
        url=request.path,
        args=request.args,
        match_info=request.match_info,
    )

    return notice


class PybrakeErrorHandler(ErrorHandler):
    def default(self, request, exception):
        """
        handles errors that have no error handlers assigned and send it to
        Airbrake.
        :param request: Request object
        :param exception: Exception object
        :return:
        """
        notifier = request.app.config.get('pybrake')
        notice = notifier.build_notice(exception)
        if request:
            notice = request_filter(request, notice)
        notifier.send_notice(notice)
        return super().default(request, exception)


def init_app(app, sqlEngine=None) -> Sanic:
    if "pybrake" in app.config:
        raise ValueError("pybrake is already injected")
    if "PYBRAKE" not in app.config:
        raise ValueError("app.config['PYBRAKE'] is not defined")

    notifier = Notifier(**app.config["PYBRAKE"])
    config = notifier.config

    app.config["pybrake"] = notifier

    # Error Handler
    app.error_handler = PybrakeErrorHandler()

    # Route Stats Monitoring
    @app.middleware("request")
    async def before_request(request):
        if not config.get("performance_stats"):
            return
        metric = RouteMetric(method=request.method, route=request.url)
        set_active_metric(metric)

    @app.middleware("response")
    async def after_request(request, resp):
        if not config.get("performance_stats"):
            return

        metric = get_active_metric()
        if metric is not None:
            metric.status_code = resp.status
            metric.content_type = resp.content_type
            metric.end_time = time.time()
            notifier.routes.notify(metric)
            set_active_metric(None)

    # Query Stats Monitoring

    if _sqla_available and sqlEngine:
        _sqla_instrument(notifier, sqlEngine)

    # TODO: Add query stats for tortoise ORM  # pylint: disable=fixme

    # Route Breakdown Monitoring
    old_html_render = response.html

    def patch_html(
            body,
            status: int = 200,
            headers: Optional[Dict[str, str]] = None,
    ) -> response.HTTPResponse:
        start_span("template")
        res = old_html_render(body, status, headers)
        end_span("template")
        return res

    response.html = patch_html

    return app


def _sqla_instrument(notifier, sqlEngine):
    event.listen(sqlEngine.sync_engine, "before_cursor_execute",
                 _before_cursor(notifier))
    event.listen(sqlEngine.sync_engine, "after_cursor_execute",
                 _after_cursor(notifier))


def _before_cursor(notifier):
    def _sqla_before_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if not notifier.config.get("performance_stats"):
            return
        start_span("sql")

    return _sqla_before_cursor_execute


def _after_cursor(notifier):
    # pylint: disable=fixme
    #  TODO: In the query stats notice add filename, function, and line
    #   number. As of present, it is not possible to obtain a complete
    #   execution traceback.
    def _sqla_after_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if not notifier.config.get("performance_stats"):
            return
        end_span("sql")
        metric = get_active_metric()
        if metric is not None:
            notifier.queries.notify(
                query=statement,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=metric.start_time,
                end_time=time.time(),
            )

    return _sqla_after_cursor_execute



import json
from datetime import datetime
from random import randrange
from time import sleep

import requests
from flask import Flask, render_template, jsonify
from flask_sqlalchemy import SQLAlchemy
from pybrake.flask import init_app

app = Flask(__name__)

app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:////tmp/test.db"
app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
db = SQLAlchemy(app)

app.config["PYBRAKE"] = dict(
    project_id="XXXXXXXXXXXX",
    project_key="XXXXXXXXXXXX",
    environment="test",
    performance_stats=True,  # False to disable APM
)

app = init_app(app)

city_list = ["austin", "pune", "santabarbara", "washington"]


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(80), unique=True, nullable=False)
    email = db.Column(db.String(120), unique=True, nullable=False)

    def __repr__(self):
        return "<User %r>" % self.username


db.create_all()


@app.route("/ping", methods=["GET"])
def ping():
    return "Pong"


@app.route("/hello/<name>", methods=["GET"])
def hello(name):
    sleep(randrange(0, 3))
    print(User.query.all())
    return render_template("hello.html", name=name)


@app.route('/')
def index():
    return "Hello, Welcome to the Weather App!"


@app.route('/date')
def getdate():
    return jsonify({
        "date": "Current date and time is: %s" % datetime.now()
    })


@app.route('/locations')
def get_location_details():
    return jsonify({
        'cities': city_list
    })


# API for weather details for a location
@app.route('/weather/<location_name>')
def get_weather_details(location_name):
    if location_name not in city_list:
        return app.response_class(
            status=400,
            response=json.dumps({
                'error': 'Not found: Location not found!'
            }),
            mimetype='application/json'
        )
    with requests.get('https://airbrake.github.io/weatherapi/weather/' + location_name) as f:
        data = f.json()
        return app.response_class(
            status=200,
            response=json.dumps(data),
            mimetype='application/json'
        )


if __name__ == '__main__':
    app.debug = True
    app.run(port=3000)



# -*- coding: utf-8 -*-
"""Error controller"""
from tg import request, expose
from weather_fullstack.lib.base import BaseController

__all__ = ['ErrorController']


class ErrorController(BaseController):
    """
    Generates error documents as and when they are required.

    The ErrorDocuments middleware forwards to ErrorController when error
    related status codes are returned from the application.

    This behaviour can be altered by changing the parameters to the
    ErrorDocuments middleware in your config/middleware.py file.

    """

    @expose('weather_fullstack.templates.error')
    def document(self, *args, **kwargs):
        """Render the error document"""
        resp = request.environ.get('tg.original_response')
        try:
            # tg.abort exposes the message as .detail in response
            message = resp.detail
        except:
            message = None

        if not message:
            message = ("We're sorry but we weren't able to process "
                       " this request.")

        values = dict(prefix=request.environ.get('SCRIPT_NAME', ''),
                      code=request.params.get('code', resp.status_int if resp else 400),
                      message=request.params.get('message', message))
        return values



import requests
from time import sleep
from random import randrange
from datetime import datetime
from bottle import run, response, Bottle
from pybrake.middleware.bottle import init_app
from bottle.ext import sqlalchemy
from sqlalchemy import create_engine, Column, Integer, Sequence, String
from sqlalchemy.ext.declarative import declarative_base

app = Bottle()

app.config['PYBRAKE'] = dict(
    project_id=999999,
    project_key='xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx',
)

Base = declarative_base()
engine = create_engine('sqlite:////tmp/test.db', echo=True)

app.install(sqlalchemy.Plugin(
    engine,  # SQLAlchemy engine created with create_engine function.
    Base.metadata,  # SQLAlchemy metadata, required only if create=True.
    keyword='db',  # Keyword used to inject session database in a route (
    # default 'db').
    create=True,  # If it is true, execute `metadata.create_all(engine)`
    # when plugin  is applied (default False).
    commit=True,  # If it is true, plugin commit changes after route is
    # executed (default True).
    use_kwargs=False  # If it is true and keyword is not defined, plugin
    # uses **kwargs argument to inject session database (default False).
))
app = init_app(app)

city_list = ["austin", "pune", "santabarbara", "washington"]


class Entity(Base):
    __tablename__ = 'entity'
    id = Column(Integer, Sequence('id_seq'), primary_key=True)
    name = Column(String(50))

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "<Entity('%d', '%s')>" % (self.id, self.name)


# API for Hello Application
@app.get('/')
def hello(db):
    sleep(randrange(0, 3))
    print(db.query(Entity).all())
    return "Hello, Welcome to the Weather App!"


# API for current server date
@app.get('/date', method=['GET'])
def getdate():
    return {
        "date": "Current date and time is: %s" % datetime.now()
    }


# API for location details
@app.get('/locations')
def get_location_details():
    return {
        'cities': city_list
    }


# API for weather details for a location
@app.get('/weather/<location_name>')
def get_weather_details(location_name):
    response.headers['Content-Type'] = 'application/json'
    response.headers['Cache-Control'] = 'no-cache'
    if location_name not in city_list:
        response.status = 400
        return {
            'error': 'Not found: Location not found!'
        }
    with requests.get(
            'https://airbrake.github.io/weatherapi/weather/' +
            location_name) as f:
        data = f.json()
        return data


if __name__ == '__main__':
    run(app, host='localhost', port=3000, debug=True, reloader=True)



# -*- coding: utf-8 -*-

#  Quickstarted Options:
#
#  sqlalchemy: True
#  auth:       sqlalchemy
#  mako:       False
#
#

# This is just a work-around for a Python2.7 issue causing
# interpreter crash at exit when trying to log an info message.
try:
    import logging
    import multiprocessing
except:
    pass

import sys
py_version = sys.version_info[:2]

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

testpkgs = [
    'WebTest >= 1.2.3',
    'nose',
    'coverage',
    'gearbox'
]

install_requires = [
    "TurboGears2 >= 2.4.3",
    "Beaker >= 1.8.0",
    "Kajiki >= 0.6.3",
    "zope.sqlalchemy >= 1.2",
    "sqlalchemy",
    "alembic",
    "repoze.who",
    "tw2.forms",
    "tgext.admin >= 0.6.1",
    "WebHelpers2"
]

if py_version != (3, 2):
    # Babel not available on 3.2
    install_requires.append("Babel")

setup(
    name='weather-fullstack',
    version='0.1',
    description='',
    author='',
    author_email='',
    url='',
    packages=find_packages(exclude=['ez_setup']),
    install_requires=install_requires,
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=testpkgs,
    extras_require={
        'testing': testpkgs
    },
    package_data={'weather_fullstack': [
        'templates/*/*',
    ]},
    message_extractors={'weather_fullstack': [
        ('**.py', 'python', None),
        ('templates/**.xhtml', 'kajiki', {'strip_text': False, 'extract_python': True}),
    ]},
    entry_points={
        'paste.app_factory': [
            'main = weather_fullstack.config.application:make_app'
        ],
        'gearbox.plugins': [
            'turbogears-devtools = tg.devtools'
        ]
    },
    zip_safe=False
)



# -*- coding: utf-8 -*-
"""WSGI application initialization for weather-fullstack."""
from weather_fullstack.config.app_cfg import base_config

__all__ = ['make_app']


def make_app(global_conf, **app_conf):
    """
    Set weather-fullstack up with the settings found in the PasteDeploy configuration
    file used.

    :param dict global_conf: The global settings for weather-fullstack
                             (those defined under the ``[DEFAULT]`` section).

    :return: The weather-fullstack application with all the relevant middleware
        loaded.

    This is the PasteDeploy factory for the weather-fullstack application.

    ``app_conf`` contains all the application-specific settings (those defined
    under ``[app:main]``.
    """
    app = base_config.make_wsgi_app(global_conf, app_conf, wrap_app=None)

    # Wrap your final TurboGears 2 application with custom middleware here

    return app



import random

from celery import Celery
from pybrake.middleware.celery import patch_celery

import pybrake

app = Celery("tasks", broker="redis://localhost")

notifier = pybrake.Notifier(
    project_id=1,
    project_key="FIXME",
    environment="celery"
)
patch_celery(notifier)


@app.task
def add(x, y):
    if random.random() < 0.5:
        raise ValueError("bad luck")
    return x + y



import time
import traceback
import typing as t
from sys import exc_info as _exc_info

import cherrypy
from cherrypy.process import plugins
from pybrake.notifier import Notifier

from .. import Notifier
from .. import RouteMetric
from ..metrics import (
    set_active as set_active_metrics,
    get_active as get_active_metrics,
    start_span,
    end_span,
)

try:
    from jinja2 import Template
except ImportError:
    _jinja_available = False
else:
    _jinja_available = True

try:
    from sqlalchemy import event
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True


def request_filter(notice):
    request = cherrypy.request
    if request is None:
        return notice
    ctx = notice["context"]
    ctx["method"] = request.method
    ctx["route"] = request.path_info

    try:
        user_addr = request.remote.ip
    except IndexError:
        user_addr = request.wsgi_environ.get('REMOTE_ADDR')
    if user_addr:
        ctx["userAddr"] = user_addr

    if request.login:
        ctx["user"] = request.login

    notice["params"]["request"] = dict(
        body=request.body.__dict__,
        cookie=dict(request.cookie),
        headers=dict(request.headers),
        request_line=request.request_line,
        url_args=request.args,
        script_name=request.script_name,
        kwargs=request.kwargs,
        query_string=dict(request.query_string),
        environ=dict(request.wsgi_environ),
    )
    return notice


class PybrakeAPM(cherrypy.Tool):
    def __init__(self):
        cherrypy.Tool.__init__(self, 'before_handler',
                               self.start_timer,
                               priority=1)

    def _setup(self, **kwargs):
        cherrypy.Tool._setup(self)
        cherrypy.request.hooks.attach('on_end_request',
                                      self.end_timer,
                                      priority=0)

    def start_timer(self, **kwargs):
        notifier = cherrypy.config.get('pybrake')
        if not notifier.config.get("performance_stats"):
            return

        metric = RouteMetric(method=cherrypy.request.method,
                             route=cherrypy.request.path_info)
        set_active_metrics(metric)

    def end_timer(self, **kwargs):
        notifier = cherrypy.config.get('pybrake')
        if not notifier.config.get("performance_stats"):
            return
        metric = get_active_metrics()
        if metric is not None:
            metric.status_code = int(cherrypy.response.status[:3]) if \
                isinstance(cherrypy.response.status, str) \
                else cherrypy.response.status
            metric.content_type = cherrypy.response.headers.get("Content-Type")
            metric.end_time = time.time()
            notifier.routes.notify(metric)
            set_active_metrics(None)


cherrypy.tools.pybrake_apm = PybrakeAPM()


class PybrakeErrorNotifier(cherrypy.Tool):
    def __init__(self):
        cherrypy.Tool.__init__(self, 'before_error_response',
                               self.error_notifier,
                               priority=1)

    def _setup(self, **kwargs):
        cherrypy.Tool._setup(self)

    def error_notifier(self, **kwargs):
        notifier = cherrypy.config.get('pybrake')
        notifier.notify(_exc_info()[1])


cherrypy.tools.pybrake_error = PybrakeErrorNotifier()


class PybrakeQueryStats(cherrypy.Tool):
    def __init__(self):
        cherrypy.Tool.__init__(self, 'on_start_resource',
                               self.start_timer,
                               priority=21)

    def _setup(self, **kwargs):
        cherrypy.Tool._setup(self)

    def start_timer(self, **kwargs):
        notifier = cherrypy.config.get('pybrake')
        if not notifier.config.get("performance_stats"):
            return

        if _sqla_available:
            event.listen(cherrypy.request.db.bind, "before_cursor_execute",
                         _before_sql_cursor(notifier.config))
            event.listen(cherrypy.request.db.bind, "after_cursor_execute",
                         _after_sql_cursor(notifier.config))


cherrypy.tools.pybrake_query_stats = PybrakeQueryStats()


def _before_sql_cursor(config):
    def _sqla_before_cursor_execute(
            conn, cursor, statement, parameters, context,
            executemany
    ):
        if not config.get("performance_stats"):
            return
        start_span("sql")

    return _sqla_before_cursor_execute


def _after_sql_cursor(config):
    def _sqla_after_cursor_execute(
            conn, cursor, statement, parameters, context,
            executemany
    ):
        notifier = cherrypy.config.get('pybrake')
        if not config.get("performance_stats"):
            return
        end_span("sql")
        metric = get_active_metrics()
        if metric is not None:
            try:
                traceback_frm = traceback.extract_stack(limit=10)[0]
            except IndexError as er:  # pylint: disable=unused-variable
                traceback_frm = None
            notifier.queries.notify(
                query=statement,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=metric.start_time,
                end_time=time.time(),
                function=traceback_frm.name if traceback_frm else '',
                file=traceback_frm.filename if traceback_frm else '',
                line=traceback_frm.lineno if traceback_frm else 0,
            )

    return _sqla_after_cursor_execute


class PybrakePlugin(plugins.SimplePlugin):
    def __init__(self, bus, project_id, project_key, **kw):
        """
        The plugin is registered to the CherryPy engine and therefore
        is part of the bus (the engine *is* a bus) registry.
        We use this plugin to create the notifier instance and apply patch
        for Error notification and APM.
        :param project_id: Airbrake project id
        :param project_key: Airbrake project key
        :param kw: Addition pybrake parameters like performance_stats,
        queue_stats, etc.
        """

        plugins.SimplePlugin.__init__(self, bus)
        self.project_id = project_id
        self.project_key = project_key
        self.kwargs = kw

        self.bus.subscribe('pybrake.bind', self.bind)

        self.notifier = None

    def start(self):
        self.notifier = Notifier(
            project_id=self.project_id,
            project_key=self.project_key,
            **self.kwargs
        )
        self.notifier.add_filter(request_filter)

        # Patch of route breakdown
        if _jinja_available:
            old_template_render = Template.render

            def patch_template_render(self, *args: t.Any, **kwargs: t.Any):
                start_span("template")
                res = old_template_render(self, *args, *kwargs)
                end_span("template")
                return res

            Template.render = patch_template_render

    def create(self):
        if 'pybrake' in cherrypy.config:
            raise ValueError("pybrake is already injected")
        self.start()
        cherrypy.config.update({
            'pybrake': self.notifier,
            'tools.pybrake_apm.on': True,
            'tools.pybrake_error.on': True,
            'tools.pybrake_query_stats.on': True,
        })

    def bind(self, session):
        session.configure(bind=self.notifier)



from warnings import warn, simplefilter

from .middleware import flask


def __getattr__(name):
    if name == 'init_app':
        simplefilter("always", DeprecationWarning)
        warn(DeprecationWarning('pybrake.flask.init_app has been '
                                'renamed to '
                                'pybrake.middleware.flask.init_app'))
        return flask.init_app
    raise AttributeError('No module named ' + name)



import time
import traceback
import hug  # pylint: disable=import-error
from ..metrics import (
    set_active as set_active_metric,
    get_active as get_active_metric,
    start_span, end_span
)
from ..notifier import Notifier
from ..route_metric import RouteMetric

try:
    from sqlalchemy import event
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True


_UNKNOWN_ROUTE = "UNKNOWN"


def request_filter(request, notice):
    if request is None:
        return notice

    ctx = notice["context"]
    ctx["method"] = request.method
    ctx["url"] = request.uri_template
    ctx["route"] = request.url

    try:
        user_addr = request.access_route[0]
    except IndexError:
        user_addr = request.remote_addr
    if user_addr:
        ctx["userAddr"] = user_addr

    if hug.directives.user:  # pylint: disable=using-constant-test
        curr_user = {}
        for s in ["username", "name"]:
            if hasattr(hug.directives.user, s):
                curr_user[s] = getattr(hug.directives.user, s)
        ctx["user"] = curr_user

    notice["params"]["request"] = dict(
        json=request.params,
        cookies=dict(request.cookies),
        headers=dict(request.headers),
        environ=request.env,
        url_rule=request.uri_template,
    )

    return notice


def before_request_middleware(notifier, req):
    if notifier and not notifier.config.get("performance_stats"):
        return

    if req.url:
        route = req.url
    else:
        route = _UNKNOWN_ROUTE

    metric = RouteMetric(method=req.method, route=route)
    set_active_metric(metric)


def after_request_middleware(notifier, req, resp):
    if notifier and not notifier.config.get("performance_stats"):
        return
    metric = get_active_metric()

    if metric is not None:
        metric.status_code = int(resp.status.split(" ")[0]) if isinstance(
            resp.status, str) else resp.status
        metric.content_type = resp.content_type
        metric.end_time = time.time()
        notifier.routes.notify(metric)
        set_active_metric(None)


def _before_sql_cursor(notifier):
    def _sqla_before_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if notifier and not notifier.config.get("performance_stats"):
            return
        start_span("sql")

    return _sqla_before_cursor_execute


def _after_sql_cursor(notifier):
    def _sqla_after_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if notifier and not notifier.config.get("performance_stats"):
            return
        end_span("sql")
        metric = get_active_metric()
        if metric is not None:
            try:
                traceback_frm = traceback.extract_stack(limit=7)[0]
            except IndexError as er:  # pylint: disable=unused-variable
                traceback_frm = None
            notifier.queries.notify(
                query=statement,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=metric.start_time,
                end_time=time.time(),
                function=traceback_frm.name if traceback_frm else '',
                file=traceback_frm.filename if traceback_frm else '',
                line=traceback_frm.lineno if traceback_frm else 0,
            )

    return _sqla_after_cursor_execute


def init_app(api, config, sqlEngine=None):
    if "pybrake" in config:
        raise ValueError("pybrake is already injected")
    if "PYBRAKE" not in config:
        raise ValueError("config['PYBRAKE'] is not defined")

    notifier = Notifier(**config['PYBRAKE'])

    # Route stats
    @hug.middleware_class(api=api)
    class PybrakeMiddleware:  # pylint: disable=unused-variable

        def __init__(self):
            self.notifier = notifier

        def process_request(self, req, resp):
            before_request_middleware(self.notifier, req=req)
            return resp

        async def process_request_async(self, req, resp):
            before_request_middleware(self.notifier, req=req)
            return resp

        def process_response(self, req, resp, resource, req_succeeded):
            after_request_middleware(notifier=self.notifier, req=req, resp=resp)
            return resp

        async def process_response_async(self, req, resp, resource,
                                         req_succeeded):
            after_request_middleware(notifier=self.notifier, req=req,
                                     resp=resp)
            return resp

    # Route Breakdown stats
    old_html = hug.output_format.html

    @hug.format.content_type("text/html; charset=utf-8")
    def _patch_html_output_format(content, **kwargs):
        start_span("template")
        res = old_html(content, *kwargs)
        end_span("template")
        return res

    hug.output_format.html = _patch_html_output_format

    # Patch Error handler
    http_call = hug.interface.HTTP.__call__

    def _patch_call__(self, request, response, api_version=None, **kwargs):
        try:
            return http_call(self, request, response, api_version=None, **kwargs)
        except Exception as e:
            notice = notifier.build_notice(e)
            notice = request_filter(request, notice)
            notifier.send_notice(notice)
            response.status = '500 Internal Server Error'
            raise e

    hug.interface.HTTP.__call__ = _patch_call__

    # Query Stats
    if sqlEngine and _sqla_available:
        event.listen(sqlEngine, "before_cursor_execute",
                     _before_sql_cursor(notifier))
        event.listen(sqlEngine, "after_cursor_execute",
                     _after_sql_cursor(notifier))

    return config



# -*- coding: utf-8 -*-
"""Setup the weather-fullstack application"""
from __future__ import print_function, unicode_literals
import transaction
from weather_fullstack import model


def bootstrap(command, conf, vars):
    """Place any commands to setup weather_fullstack here"""

    # <websetup.bootstrap.before.auth
    from sqlalchemy.exc import IntegrityError
    try:
        u = model.User()
        u.user_name = 'manager'
        u.display_name = 'Example manager'
        u.email_address = 'manager@somedomain.com'
        u.password = 'managepass'

        model.DBSession.add(u)

        g = model.Group()
        g.group_name = 'managers'
        g.display_name = 'Managers Group'

        g.users.append(u)

        model.DBSession.add(g)

        p = model.Permission()
        p.permission_name = 'manage'
        p.description = 'This permission gives an administrative right'
        p.groups.append(g)

        model.DBSession.add(p)

        u1 = model.User()
        u1.user_name = 'editor'
        u1.display_name = 'Example editor'
        u1.email_address = 'editor@somedomain.com'
        u1.password = 'editpass'

        model.DBSession.add(u1)
        model.DBSession.flush()
        transaction.commit()
    except IntegrityError:
        print('Warning, there was a problem adding your auth data, '
              'it may have already been added:')
        import traceback
        print(traceback.format_exc())
        transaction.abort()
        print('Continuing with bootstrapping...')

    # <websetup.bootstrap.after.auth>



# -*- coding: utf-8 -*-
"""
Global configuration file for TG2-specific settings in weather-fullstack.

This file complements development/deployment.ini.

"""
from tg import FullStackApplicationConfigurator

import weather_fullstack
from weather_fullstack import model, lib
from pybrake.middleware.turbogears import init_app

base_config = FullStackApplicationConfigurator()

# General configuration
base_config.update_blueprint({
    # True to prevent dispatcher from striping extensions
    # For example /socket.io would be served by "socket_io"
    # method instead of "socket".
    'disable_request_extensions': False,

    # Set None to disable escaping punctuation characters to "_"
    # when dispatching methods.
    # Set to a function to provide custom escaping.
    'dispatch_path_translator': True,

    'package': weather_fullstack,
})

# ToscaWidgets configuration
base_config.update_blueprint({
    'tw2.enabled': True,
})

# Rendering Engines Configuration
rendering_config = {
    'renderers': ['json'],  # Enable json in expose
    'default_renderer': 'kajiki',
}
rendering_config['renderers'].append('kajiki')
# Change this in setup.py too for i18n to work.
rendering_config['templating.kajiki.strip_text'] = False
base_config.update_blueprint(rendering_config)

# Configure Sessions, store data as JSON to avoid pickle security issues
base_config.update_blueprint({
    'session.enabled': True,
    'session.data_serializer': 'json',
})

# Configure the base SQLALchemy Setup
base_config.update_blueprint({
    'use_sqlalchemy': True,
    'model': weather_fullstack.model,
    'DBSession': weather_fullstack.model.DBSession,
})
# This tells to TurboGears how to retrieve the data for your user
from tg.configuration.auth import TGAuthMetadata
class ApplicationAuthMetadata(TGAuthMetadata):
    def __init__(self, dbsession, user_class):
        self.dbsession = dbsession
        self.user_class = user_class

    def authenticate(self, environ, identity):
        login = identity['login']
        user = self.dbsession.query(self.user_class).filter_by(
            user_name=login
        ).first()

        if not user:
            login = None
        elif not user.validate_password(identity['password']):
            login = None

        if login is None:
            try:
                from urllib.parse import parse_qs, urlencode
            except ImportError:
                from urlparse import parse_qs
                from urllib import urlencode
            from tg.exceptions import HTTPFound

            params = parse_qs(environ['QUERY_STRING'])
            params.pop('password', None)  # Remove password in case it was there
            if user is None:
                params['failure'] = 'user-not-found'
            else:
                params['login'] = identity['login']
                params['failure'] = 'invalid-password'

            # When authentication fails send user to login page.
            environ['repoze.who.application'] = HTTPFound(
                location=environ['SCRIPT_NAME'] + '?'.join(('/login', urlencode(params, True)))
            )

        return login

    def get_user(self, identity, userid):
        return self.dbsession.query(self.user_class).filter_by(
            user_name=userid
        ).first()

    def get_groups(self, identity, userid):
        return [g.group_name for g in identity['user'].groups]

    def get_permissions(self, identity, userid):
        return [p.permission_name for p in identity['user'].permissions]

# Configure the authentication backend
base_config.update_blueprint({
    'auth_backend': 'sqlalchemy',

    # YOU MUST CHANGE THIS VALUE IN PRODUCTION TO SECURE YOUR APP
    'sa_auth.cookie_secret': "f9a6a628-ef0c-4a85-aa73-62a449e669f9",
    'sa_auth.authmetadata': ApplicationAuthMetadata(model.DBSession, model.User),

    # You may optionally define a page where you want users
    # to be redirected to on login:
    'sa_auth.post_login_url': '/post_login',

    # You may optionally define a page where you want users
    # to be redirected to on logout:
    'sa_auth.post_logout_url': '/post_logout',
    
    # In case ApplicationAuthMetadata didn't find the user discard the whole identity.
    # This might happen if logged-in users are deleted.
    'identity.allow_missing_user': False,

    # override this if you would like to provide a different who plugin for
    # managing login and logout of your application
    'sa_auth.form_plugin': None,
    
    # You can use a different repoze.who Authenticator if you want to
    # change the way users can login
    # 'sa_auth.authenticators': [('myauth', SomeAuthenticator()],
    
    # You can add more repoze.who metadata providers to fetch
    # user metadata.
    # Remember to set 'sa_auth.authmetadata' to None
    # to disable authmetadata and use only your own metadata providers
    # 'sa_auth.mdproviders': [('myprovider', SomeMDProvider()],
})
try:
    # Enable DebugBar if available, install tgext.debugbar to turn it on
    from tgext.debugbar import enable_debugbar
    enable_debugbar(base_config)
except ImportError:
    pass

base_config.update_blueprint({
    'PYBRAKE': {
        "project_id": 99999,
        "project_key": "xxxxxxxxxxxxxxxxx",
    },
})

base_config = init_app(base_config)



import time

import pytest
import pybrake.metrics as metrics
from pybrake.route_metric import RouteMetric, _RouteBreakdown, RouteBreakdowns
from pybrake.routes import _Routes

metrics.FLUSH_PERIOD = 0
CONFIG = {
    "config": {
        "performance_stats": False,
        "backlog_enabled": True,
        "error_host": "http://localhost:5000",
        "apm_host": "http://localhost:5000",
    }
}


def _test_setup(config: dict, status_code=200):
    routes = RouteBreakdowns(**config)

    metric = RouteMetric(method="GET", route="/test")
    metric.status_code = status_code
    metric.content_type = "application/json"
    metric.FLUSH_PERIOD = 5
    metric._groups = {'route': 24.0, 'test': 0.4}
    metric.end_time = time.time()

    key = metric._key()
    metric.end()
    stat = _RouteBreakdown(
        method=metric.method, route=metric.route,
        responseType="application/json",
        time=1648551580.0367732)
    routes._stats = {key: stat}
    stat.add((metric.end_time - metric.start_time) * 1000)
    return routes, metric


def test_routes_breakdowns_flash_empty_stats():
    routes = _Routes(**CONFIG)

    metric = RouteMetric(method="GET", route="/test")
    metric.status_code = 200
    metric.content_type = "application/json"
    metric.end_time = time.time()
    with pytest.raises(ValueError, match=r"stats is empty"):
        routes.breakdowns._env = "Test"
        routes.breakdowns._flush()


def test_routes_breakdowns_notify_flush(mocker):
    mocker.patch(
        "pybrake.metrics.send",
        return_value=None
    )
    routes, metric = _test_setup(config=CONFIG)
    assert routes._flush() is None


def test_routes_breakdowns_notify(mocker):
    mocker.patch(
        "pybrake.route_metric.RouteBreakdowns._flush",
        return_value=None
    )
    routes, metric = _test_setup(config=CONFIG)
    assert routes.notify(metric) is None


def test_routes_breakdowns_notify_with_performance(mocker):
    mocker.patch(
        "pybrake.route_metric.RouteBreakdowns._flush",
        return_value=None
    )
    CONFIG.get('config').update({
        'performance_stats': True,
    })
    routes, metric = _test_setup(config=CONFIG)
    routes._stats = None
    assert routes.notify(metric) is None


def test_routes_breakdowns_notify_response_500(mocker):
    mocker.patch(
        "pybrake.route_metric.RouteBreakdowns._flush",
        return_value=None
    )
    CONFIG.get('config').update({
        'performance_stats': True,
    })
    routes, metric = _test_setup(config=CONFIG, status_code=500)
    routes._stats = None
    assert routes.notify(metric) is None


def test_routes_breakdowns_notify_response_400(mocker):
    mocker.patch(
        "pybrake.route_metric.RouteBreakdowns._flush",
        return_value=None
    )
    CONFIG.get('config').update({
        'performance_stats': True,
    })
    routes, metric = _test_setup(config=CONFIG, status_code=400)
    routes._stats = None
    assert routes.notify(metric) is None


def test_initiate_route_breakdown():
    route = _RouteBreakdown(
        method="GET", route="/test",
        responseType="application/json",
        time=1648551580.0367732)
    assert route.__dict__ == {'count': 0,
                              'sum': 0,
                              'sumsq': 0,
                              'tdigest': 'AAAAAkAkAAAAAAAAAAAAAA==',
                              'groups': {},
                              'method': 'GET',
                              'route': '/test',
                              'responseType': 'application/json',
                              'time': '2022-03-29T10:59:00Z'}


def test_routes_ab_url():
    CONFIG.update({'project_id': 0})
    stats = RouteBreakdowns(**CONFIG)
    assert stats._ab_url() == \
           'http://localhost:5000/api/v5/projects/0/routes-breakdowns'



import logging
import time
import traceback

from morepath.authentication import NoIdentity
from morepath.directive import HtmlAction
from morepath.view import View

from morepath import render_json, app as application, render_html
from ..metrics import (
    set_active as set_active_metric,
    get_active as get_active_metric,
    start_span, end_span)
from ..notifier import Notifier
from ..route_metric import RouteMetric

try:
    from sqlalchemy import event
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True

_UNKNOWN_ROUTE = "UNKNOWN"

logger = logging.getLogger(__name__)


# Route breakdown
def renderer(function):
    def wrapper(*args, **kw):
        start_span("template")
        res = function(*args, **kw)
        end_span("template")
        return res

    return wrapper


def patch__init__(self, model, render=None, template=None, load=None,
                  permission=None, internal=False, **predicates):
    render = render or renderer(render_html)
    super(HtmlAction, self).__init__(
        model, render, template, load, permission, internal, **predicates
    )


HtmlAction.__init__ = patch__init__


# Route Stats Monitoring
old__call__view = View.__call__

def __patch_call__(self, app, obj, request):
    notifier = app.settings.pybrake
    before_request(request, notifier.config)
    try:
        res = old__call__view(self, app, obj, request)
        after_request(notifier, res)
        return res
    except Exception as er:

        raise er

View.__call__ = __patch_call__


def before_request(request, config):
    # pylint: disable=fixme
    # TODO: When there is a path variable, pass the route url as a dynamic
    #  one. Ex:- /weather/{location_name} rather then /weather/pune
    if not config.get("performance_stats"):
        return
    metric = RouteMetric(method=request.method, route=request.path)
    set_active_metric(metric)


def after_request(notifier, resp):
    if not notifier.config.get("performance_stats"):
        return
    metric = get_active_metric()
    if metric is not None:
        metric.status_code = resp.status_code
        metric.content_type = resp.content_type
        metric.end_time = time.time()
        notifier.routes.notify(metric)
        set_active_metric(None)


# Error Handler
def request_filter(request, notice):
    if request is None:
        return notice
    ctx = notice["context"]
    ctx["method"] = request.method
    ctx["url"] = request.url
    ctx["path_code_info"] = str(request.path_code_info.sourceline)
    ctx["userAddr"] = request.remote_addr

    user = {}
    try:
        if request.identity and not isinstance(request.identity, NoIdentity):
            user_data = request.identity.as_dict()
            for s in ["userid", "username", "name"]:
                user[s] = user_data.get(s, None)
            ctx["user"] = user
    except AttributeError:
        ctx["user"] = user

    notice["params"]["request"] = dict(
        body=dict(request.body),
        params=dict(request.params) if request.params else {},
        cookies=dict(request.cookies),
        headers=dict(request.headers),
        url=request.path,
        env=request.environ,
    )

    return notice


def handle_exception(notifier, request, exception):
    notice = notifier.build_notice(exception)
    if request:
        notice = request_filter(request, notice)
    notifier.send_notice(notice)


@application.App.tween_factory()
def report_error(app, handler):
    def tween(request):
        try:
            response = handler(request)
            return response
        except Exception as er:  # pylint: disable=broad-except
            logger.exception(er)
            handle_exception(app.settings.pybrake, request, er)
            response = render_json(
                {
                    "app": repr(type(request.app)),
                    "unconsumed": request.unconsumed,
                    "error": str(er.__class__) + ':' + str(er),
                },
                request,
            )
            response.status_code = 500
            after_request(app.settings.pybrake, response)
            return response

    return tween


def init_app(app, sqlEngine=None):
    if getattr(app.settings, 'pybrake', None):
        raise ValueError("pybrake is already injected")
    if not getattr(app.settings, 'PYBRAKE', None):
        raise ValueError("app.settings.PYBRAKE is not defined")

    notifier = Notifier(**app.settings.PYBRAKE.__dict__)

    app.settings.pybrake = notifier

    # Query Stats Monitoring
    if _sqla_available and sqlEngine:
        _sqla_instrument(notifier, sqlEngine)

    return app


def _sqla_instrument(notifier, sqlEngine):
    event.listen(sqlEngine, "before_cursor_execute",
                 _before_cursor(notifier))
    event.listen(sqlEngine, "after_cursor_execute",
                 _after_cursor(notifier))


def _before_cursor(notifier):
    def _sqla_before_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if not notifier.config.get("performance_stats"):
            return
        start_span("sql")

    return _sqla_before_cursor_execute


def _after_cursor(notifier):
    def _sqla_after_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if not notifier.config.get("performance_stats"):
            return
        end_span("sql")
        metric = get_active_metric()
        if metric is not None:
            try:
                traceback_frm = traceback.extract_stack(limit=7)[0]
            except IndexError as er:  # pylint: disable=unused-variable
                traceback_frm = None
            notifier.queries.notify(
                query=statement,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=metric.start_time,
                end_time=time.time(),
                function=traceback_frm.name if traceback_frm else '',
                file=traceback_frm.filename if traceback_frm else '',
                line=traceback_frm.lineno if traceback_frm else 0,
            )

    return _sqla_after_cursor_execute



import functools
import traceback
import time

from masonite.middleware import Middleware
from masonite.views import View

from .. import Notifier
from .. import QueueMetric
from .. import RouteMetric
from ..metrics import (
    set_active as set_active_metric,
    get_active as get_active_metric,
    start_span,
    end_span,
)

try:
    from masoniteorm.connections.BaseConnection import BaseConnection
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True


def request_filter(request, notice):
    if request is None:
        return notice
    ctx = notice["context"]
    ctx["method"] = request.get_request_method()
    ctx["url"] = request.route.url
    ctx["route"] = str(request.route)
    ctx["userAddr"] = request.ip()

    user = {}
    try:
        if request.user():
            for s in ["username", "name", "email"]:
                if hasattr(request.user(), s):
                    user[s] = getattr(request.user(), s)
            ctx["user"] = user
    except AttributeError:
        ctx["user"] = user

    notice["params"]["request"] = dict(
        path_with_query=request.get_path_with_query,
        input=dict(request.all()),
        headers=request.header_bag.to_dict(),
        url=request.get_path(),
    )

    return notice


class PybrakeNotifier:

    def __init__(self, application):
        self.application = application

        if self.application.objects.get('pybrake') and \
                self.application.objects.get('pybrake').notifier:
            raise ValueError("pybrake is already injected")
        if not self.application.objects.get('config').has(
                'application.pybrake'):
            raise ValueError("app.config['PYBRAKE'] is not defined")

        self.notifier = Notifier(**self.application.objects.get('config').get(
            'application.pybrake'))
        self.application.objects['pybrake'] = self.notifier

        # Route Breakdown
        old_render = View.render

        def patch_render(selfC, template: str, dictionary: dict = None) -> \
                "View":
            start_span('template')
            res = old_render(selfC, template, dictionary)
            end_span('template')
            return res

        View.render = patch_render

        # Query Stats
        if _sqla_available:
            old_statement = BaseConnection.statement

            def patch_statement(selfC, query, bindings=()):
                if self.notifier.config.get('performance_stats'):
                    start_span("sql")
                res = old_statement(selfC, query, bindings)
                if self.notifier.config.get('performance_stats'):
                    end_span("sql")
                    metric = get_active_metric()
                    if metric is not None:
                        try:
                            traceback_frm = traceback.extract_stack(limit=5)[0]
                        except IndexError as er:  # pylint: disable=unused-variable
                            traceback_frm = None
                        self.notifier.queries.notify(
                            query=query,
                            method=getattr(metric, "method", ""),
                            route=getattr(metric, "route", ""),
                            start_time=metric.start_time,
                            end_time=time.time(),
                            function=traceback_frm.name if traceback_frm else '',
                            file=traceback_frm.filename if traceback_frm else '',
                            line=traceback_frm.lineno if traceback_frm else 0,
                        )
                return res

            BaseConnection.statement = patch_statement


# Error Notification
class PybrakeErrorListener:

    def handle(self, exception_type, exception):
        try:
            # pylint: disable=import-outside-toplevel
            from wsgi import application
        except ImportError:
            application = False
        if not application:
            return None
        request = application.make('request')
        notice = request.app.objects['pybrake'].notifier.build_notice(
            exception)
        if request:
            notice = request_filter(request, notice)
        request.app.objects['pybrake'].notifier.send_notice(notice)
        return None


# Route Stats
class PybrakeRouteMiddleware(Middleware):
    def before(self, request, response):
        if not request.app.objects['pybrake'].notifier.config.get(
                "performance_stats"):
            return None
        metric = RouteMetric(method=request.get_request_method(),
                             route=request.route.url)
        set_active_metric(metric)

        return request

    def after(self, request, response):
        notifier = request.app.objects['pybrake'].notifier
        if not notifier.config.get("performance_stats"):
            return
        metric = get_active_metric()
        if metric is not None:
            metric.status_code = response.get_status() or 500
            metric.content_type = response.header_bag.to_dict().get(
                'Content-Type')
            metric.end_time = time.time()
            notifier.routes.notify(metric)
            set_active_metric(None)


# Queue Stats
def schedule_task(fn):
    @functools.wraps(fn)
    def wrap(*args, **kwargs):
        try:
            # pylint: disable=import-outside-toplevel
            from wsgi import application
        except ImportError:
            application = False
        if application:
            metric = QueueMetric(queue=args[0].name)
            set_active_metric(metric)
            notifier = application.objects['pybrake'].notifier
            try:
                return fn(*args, **kwargs)
            except Exception as exc:
                notifier.notify(exc)
                raise exc
            finally:
                notifier.queues.notify(metric)
                set_active_metric(None)
        else:
            return fn(*args, **kwargs)

    return wrap



from masonite.middleware import Middleware


class AuthenticationMiddleware(Middleware):
    """Middleware to check if the user is logged in."""

    def before(self, request, response):
        if not request.user():
            return response.redirect(name="login")
        return request

    def after(self, request, response):
        return request



import time
import traceback

from pycnic.core import WSGI

from ..metrics import (
    set_active as set_active_metric,
    get_active as get_active_metric,
    start_span, end_span)
from ..notifier import Notifier
from ..route_metric import RouteMetric

try:
    from sqlalchemy import event
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True

try:
    from jinja2 import Template
except ImportError:
    pass
else:
    # Patch of route breakdown
    old_template_render = Template.render


    def patch_template_render(self, *args, **kwargs):
        start_span("template")
        res = old_template_render(self, *args, *kwargs)
        end_span("template")
        return res


    Template.render = patch_template_render

_UNKNOWN_ROUTE = "UNKNOWN"


# Route Stats Monitoring
def before_request(handler):
    if not handler.notifier.config.get("performance_stats"):
        return
    metric = RouteMetric(method=handler.request.method,
                         route=handler.request.path)
    set_active_metric(metric)


def after_request(handler):
    if not handler.notifier.config.get("performance_stats"):
        return

    metric = get_active_metric()
    if metric is not None:
        metric.status_code = handler.response.status_code
        metric.content_type = handler.response.header_dict.get('Content-Type')
        metric.end_time = time.time()
        handler.notifier.routes.notify(metric)
        set_active_metric(None)


def _before_sql_cursor(notifier):
    def _sqla_before_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if notifier and not notifier.config.get("performance_stats"):
            return
        start_span("sql")

    return _sqla_before_cursor_execute


def _after_sql_cursor(notifier):
    def _sqla_after_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if notifier and not notifier.config.get("performance_stats"):
            return
        end_span("sql")
        metric = get_active_metric()
        if metric is not None:
            try:
                traceback_frm = traceback.extract_stack(limit=7)[0]
            except IndexError as er:  # pylint: disable=unused-variable
                traceback_frm = None
            notifier.queries.notify(
                query=statement,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=metric.start_time,
                end_time=time.time(),
                function=traceback_frm.name if traceback_frm else '',
                file=traceback_frm.filename if traceback_frm else '',
                line=traceback_frm.lineno if traceback_frm else 0,
            )

    return _sqla_after_cursor_execute


def request_filter(request, notice):
    if request is None:
        return notice

    ctx = notice["context"]
    ctx["method"] = request.method
    ctx["url"] = request.path
    try:
        ctx["userAddr"] = request.ip
    except IndexError:
        ctx["userAddr"] = request.remote_addr

    notice["params"]["request"] = dict(
        json=request.json_args,
        cookies=dict(request.cookies),
        headers=dict(request.headers),
        environ=request.environ,
        url_rule=request.path,
    )

    return notice


class PybrakeEnableWSGI(WSGI):
    # Route Stats Monitoring
    before = before_request
    after = after_request

    def __init__(self, environ, start_response):
        if "pybrake" in getattr(self, "config"):
            raise ValueError("pybrake is already injected")
        if "PYBRAKE" not in getattr(self, "config"):
            raise ValueError("app.config['PYBRAKE'] is not defined")

        self.notifier = Notifier(**getattr(self, "config")["PYBRAKE"])
        if _sqla_available and getattr(self, "sqlDBEngine"):
            event.listen(getattr(self, "sqlDBEngine"), "before_cursor_execute",
                         _before_sql_cursor(self.notifier))
            event.listen(getattr(self, "sqlDBEngine"), "after_cursor_execute",
                         _after_sql_cursor(self.notifier))
        super().__init__(environ, start_response)

    # Error notice
    def delegate(self):
        try:
            return super().delegate()
        except AttributeError as e:
            raise e
        except Exception as err:
            notice = self.notifier.build_notice(err)
            notice = request_filter(self.request, notice)
            self.notifier.send_notice(notice)
            raise err



from datetime import *
from random import randrange
from time import sleep
from wsgiref.simple_server import make_server

import requests
from pybrake.middleware.pyramid import init_pybrake_config
from pyramid.config import Configurator
from pyramid.response import Response
from pyramid.view import view_config
from pyramid_basemodel import Base, BaseMixin
from sqlalchemy import Column, String

city_list = ["pune", "austin", "santabarbara", "washington"]

PYBRAKE_CONFIG = dict(
    project_id=999999,
    project_key='xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx',
)


class User(Base, BaseMixin):
    __tablename__ = "user"
    name = Column(String(80), unique=True, nullable=False)
    email = Column(String(120), unique=True, nullable=False)

    def __repr__(self):
        return "<User %r>" % self.name


# API for Hello Application
@view_config(route_name='hello')
def hello(request):
    sleep(randrange(0, 3))
    print(User.query.all())
    return Response('Hello, Welcome to the Weather App!')


# API for current server date
@view_config(route_name='date', renderer='json')
def date(request):
    return {
        'date': 'Current Date and Time is: %s' % datetime.now()
    }


# API for location details
@view_config(route_name='locations', renderer='json')
def locations(request):
    return {
        'locations': city_list
    }


# API for weather details for a location
@view_config(route_name='weather', renderer='json')
def weather(request):
    location_name = request.matchdict.get('location_name')
    if location_name not in city_list:
        request.response.status = 400
        return {
            'errors': {
                'message': 'Location not found!'
            }
        }
    if location_name == 'washington':
        5 / 0
    with requests.get(
            'https://airbrake.github.io/weatherapi/weather/' + location_name) as f:
        return f.json()


if __name__ == '__main__':
    settings = {
        'reload_all': True,
        'debug_all': True,
        'PYBRAKE': PYBRAKE_CONFIG,
        'sqlalchemy.url': 'sqlite:///:memory:',
        'basemodel.should_create_all': True,
        'fixtures': True,
    }
    with Configurator(settings=settings) as config:
        config.include('pyramid_basemodel')
        config.include('pyramid_tm')
        config.add_route('hello', '/')
        config.add_route('date', 'date')
        config.add_route('locations', 'locations')
        config.add_route('weather', 'weather/{location_name}')
        config.scan(".")

        config = init_pybrake_config(config)
        app = config.make_wsgi_app()

    server = make_server('0.0.0.0', 3000, app)
    server.serve_forever()



import time
import traceback

from tornado.web import RequestHandler, HTTPError

from ..metrics import (
    set_active as set_active_metrics,
    get_active as get_active_metrics,
    start_span,
    end_span,
)
from ..notifier import Notifier
from ..route_metric import RouteMetric

try:
    import tornado_sqlalchemy as _
    from sqlalchemy import event
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True

_UNKNOWN_ROUTE = "UNKNOWN"


def request_filter(notice, handler):
    request = handler.request
    if request is None:
        return notice

    ctx = notice["context"]
    ctx["method"] = request.method
    ctx["url"] = request.uri
    ctx["route"] = str(request.host)

    if request.remote_ip:
        ctx["userAddr"] = request.remote_ip

    if handler.current_user:
        user = dict(id=getattr(handler.current_user, 'id'))
        for s in ["username", "name"]:
            if hasattr(handler.current_user, s):
                user[s] = getattr(handler.current_user, s)
        ctx["user"] = user

    notice["params"]["request"] = dict(
        body_arguments=dict(request.body_arguments),
        body=dict(request.body),
        files=dict(request.files),
        cookies=dict(request.cookies),
        headers=dict(request.headers),
        query_arguments=request.query_arguments,
        path_kwargs=handler.path_kwargs,
    )

    return notice


def init_app(app):
    if "pybrake" in app.settings:
        raise ValueError("pybrake is already injected")
    if "PYBRAKE" not in app.settings:
        raise ValueError("app.settings['PYBRAKE'] is not defined")

    notifier = Notifier(**app.settings["PYBRAKE"])
    config = notifier.config

    app.settings["pybrake"] = notifier

    # Exception handling
    old_log_exception = RequestHandler.log_exception

    def _patch_log_exception(self, ty, value, tb):
        if not isinstance(value, HTTPError):
            _handle_exception(notifier, value, self)
        return old_log_exception(self, ty, value, tb)

    RequestHandler.log_exception = _patch_log_exception

    # Route stats
    old_execute = RequestHandler._execute

    async def _patch_execute(self, transforms, *args, **kwargs):
        if config.get("performance_stats"):
            _before_request(self.request, self)
        res = await old_execute(self, transforms, *args, **kwargs)
        if config.get("performance_stats"):
            _after_request(self, notifier)
        return res

    RequestHandler._execute = _patch_execute

    # Route Breakdown stats
    old_render = RequestHandler.render

    def _patch_render(self, template_name, **kwargs):
        start_span("template")
        res = old_render(self, template_name, **kwargs)
        end_span("template")
        return res

    RequestHandler.render = _patch_render

    if _sqla_available:
        _sqla_instrument(config, app)

    return app


def _before_request(request, app):
    request_route = None
    for rule in app.application.wildcard_router.rules:
        match = rule.matcher.match(request)
        if match is not None:
            request_route = rule.matcher.regex.pattern

    for rule in app.application.default_router.rules:
        match = rule.matcher.match(request)
        if match is not None and not request_route:
            request_route = rule.matcher.regex.pattern

    if request_route:
        if request_route.endswith("$"):
            request_route = request_route[:len(request_route) - 1]
        route = request_route
    elif request.url:
        route = request.uri
    else:
        route = _UNKNOWN_ROUTE
    metric = RouteMetric(method=request.method, route=route)
    set_active_metrics(metric)


def _after_request(response, notifier):
    metric = get_active_metrics()
    if metric is not None:
        metric.status_code = response.get_status()
        if 'Content-Type' in response._headers:
            metric.content_type = response._headers.get('Content-Type')
        metric.end_time = time.time()
        notifier.routes.notify(metric)
        set_active_metrics(None)


def _before_render_template(sender, template, context, **extra):
    start_span("template")


def _template_rendered(sender, template, context, **extra):
    end_span("template")


def _handle_exception(notifier, exception, handler):
    notice = notifier.build_notice(exception)
    notice = request_filter(notice, handler)
    notifier.send_notice(notice)


def _sqla_instrument(config, app):
    sqla = app.settings.get('db')
    event.listen(sqla.engine, "before_cursor_execute", _before_cursor(config))
    event.listen(sqla.engine, "after_cursor_execute", _after_cursor(
        config, app))


def _before_cursor(config):
    def _sqla_before_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if not config.get("performance_stats"):
            return
        start_span("sql")

    return _sqla_before_cursor_execute


def _after_cursor(config, app):
    def _sqla_after_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
    ):
        if not config.get("performance_stats"):
            return
        end_span("sql")
        metric = get_active_metrics()
        if metric is not None:
            try:
                traceback_frm = traceback.extract_stack(limit=12)[0]
            except IndexError as er:  # pylint: disable=unused-variable
                traceback_frm = None
            app.settings["pybrake"].queries.notify(
                query=statement,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=metric.start_time,
                end_time=time.time(),
                function=traceback_frm.name if traceback_frm else '',
                file=traceback_frm.filename if traceback_frm else '',
                line=traceback_frm.lineno if traceback_frm else 0,
            )

    return _sqla_after_cursor_execute



"""Task Module Description"""
import time
from masonite.scheduling import Task
from pybrake.middleware.masonite import schedule_task


class WeatherTest(Task):

    name = "WeatherTest"

    @schedule_task
    def handle(self):
        time.sleep(10)



import json
import threading
import time as pytime
import urllib.error
import urllib.request
from contextlib import contextmanager

from .constant import MaxRetryAttempt
from .notice import jsonify_notice
from .utils import logger

threadLocal = threading.local()

_METRIC_KEY = "_ab_metric"

_STATUS_CODE_CRITERIA_FOR_BACKLOG = (404, 408, 409, 410, 500, 502, 504)

APM_Backlog = None
Error_Backlog = None


@contextmanager
def activated_metric(metric):
    set_active(metric)
    yield
    set_active(None)


def set_active(metric):
    setattr(threadLocal, _METRIC_KEY, metric)


def get_active():
    return getattr(threadLocal, _METRIC_KEY, None)


def start_span(name, **kwargs):
    metric = get_active()
    if metric is not None:
        metric.start_span(name, **kwargs)


def end_span(name, **kwargs):
    metric = get_active()
    if metric is not None:
        metric.end_span(name, **kwargs)


class Metric:
    def __init__(self):
        self.start_time = pytime.time()
        self.end_time = None

        self._spans = {}
        self._curr_span = None

        self._groups = {}

    def end(self):
        if self.end_time is None:
            self.end_time = pytime.time()

    def _new_span(self, name, **kwargs):
        return Span(metric=self, name=name, **kwargs)

    def start_span(self, name, *, start_time=None):
        if self._curr_span is not None:
            if self._curr_span.name == name:
                self._curr_span._level += 1
                return
            self._curr_span._pause()

        span = self._spans.get(name)
        if span is None:
            span = self._new_span(name, start_time=start_time)
            self._spans[name] = span
        else:
            span._resume()

        span._parent = self._curr_span
        self._curr_span = span

    def end_span(self, name, *, end_time=None):
        if self._curr_span is not None and self._curr_span.name == name:
            if self._end_span(self._curr_span):
                self._curr_span = self._curr_span._parent
                if self._curr_span is not None:
                    self._curr_span._resume()
            return

        span = self._spans.get(name)
        if span is None:
            logger.error("pybrake: span=%s does not exist", name)
            return
        self._end_span(span, end_time=end_time)

    def _end_span(self, span, *, end_time=None):
        if span._level > 0:
            span._level -= 1
            return False

        span.end(end_time=end_time)
        self._spans.pop(span.name)
        return True

    def _inc_group(self, name, ms):
        self._groups[name] = self._groups.get(name, 0) + ms


class Span:
    def __init__(self, *, metric=None, name="", start_time=None):
        self._metric = metric
        self._parent = None

        self.name = name
        if start_time is not None:
            self.start_time = start_time
        else:
            self.start_time = pytime.time()
        self.end_time = None

        self._dur = 0
        self._level = 0

    def end(self, end_time=None):
        if end_time is not None:
            self.end_time = end_time
        else:
            self.end_time = pytime.time()
        self._dur += (self.end_time - self.start_time) * 1000
        self._metric._inc_group(self.name, self._dur)
        self._metric = None

    def _pause(self):
        if self._paused():
            return
        self._dur += (pytime.time() - self.start_time) * 1000
        self.start_time = 0

    def _paused(self):
        return self.start_time == 0

    def _resume(self):
        if not self._paused():
            return
        self.start_time = pytime.time()


def send(url, headers, payload=None, method=None, retry_count=0):
    if payload is None:
        payload = {}

    req = urllib.request.Request(
        url, data=payload, headers=headers, method=method
    )

    try:
        resp = urllib.request.urlopen(req, timeout=5)
    except urllib.error.HTTPError as err:
        resp = err
    except Exception as err:  # pylint: disable=broad-except
        logger.error(err)
        return

    try:
        body = resp.read()
    except IOError as err:
        logger.error(err)
        return

    if 200 <= resp.code < 300:
        return

    if resp.code in _STATUS_CODE_CRITERIA_FOR_BACKLOG and APM_Backlog is not\
            None and retry_count < MaxRetryAttempt:
        APM_Backlog.append_stats(val=payload, url=url, retry_count=retry_count)

    if not 400 <= resp.code < 500:
        err = f"airbrake: unexpected response status_code={resp.code}"
        logger.error(err)
        return

    if resp.code == 429:
        return

    try:
        body = body.decode("utf-8")
    except UnicodeDecodeError as err:
        logger.error(err)
        return

    try:
        in_data = json.loads(body)
    except ValueError as err:  # json.JSONDecodeError requires Python 3.5+
        logger.error(err)
        return

    if "message" in in_data:
        logger.error(in_data["message"])
        return


def send_notice(notifier, notice, url, headers, method=None, retry_count=0):
    payload = jsonify_notice(notice)
    req = urllib.request.Request(
        url, data=payload, headers=headers, method=method)

    try:
        resp = urllib.request.urlopen(req, timeout=5)
    except urllib.error.HTTPError as err:
        resp = err
    except Exception as err:  # pylint: disable=broad-except
        notice["error"] = err
        logger.error(notice["error"])

        return notice

    try:
        body = resp.read()
    except IOError as err:
        notice["error"] = err
        logger.error(notice["error"])
        return notice

    if resp.code in _STATUS_CODE_CRITERIA_FOR_BACKLOG and Error_Backlog is \
            not None and retry_count < MaxRetryAttempt:
        Error_Backlog.append_stats(val=notice, url=url, retry_count=retry_count)

    if not (200 <= resp.code < 300 or 400 <= resp.code < 500):
        notice["error"] = f"airbrake: unexpected response " \
                          f"status_code={resp.code}"
        notice["error_info"] = dict(code=resp.code, body=body)
        logger.error(notice["error"])
        return notice

    if resp.code == 429:
        return notifier._rate_limited(notice, resp)

    try:
        body = body.decode("utf-8")
    except UnicodeDecodeError as err:
        notice["error"] = err
        logger.error(notice["error"])
        return notice

    try:
        data = json.loads(body)
    except ValueError as err:  # json.JSONDecodeError requires Python 3.5+
        notice["error"] = err
        logger.error(notice["error"])
        return notice

    if "id" in data:
        notice["id"] = data["id"]
        return notice

    if "message" in data:
        notice["error"] = data["message"]
        logger.error(notice["error"])
        return notice

    notice["error"] = "unexpected Airbrake response"
    notice["error_info"] = dict(data=data)
    logger.error(notice["error"])
    return notice



import os
import asyncio
import requests
from datetime import datetime

import tornado.web
from tornado_sqlalchemy import SQLAlchemy
from tornado_sqlalchemy import SessionMixin
from pybrake.middleware.tornado import init_app

from sqlalchemy import Column, Integer, String

city_list = ["pune", "austin", "santabarbara", "washington"]

# SQL setup
db = SQLAlchemy(url="sqlite:///:memory:")


class User(db.Model):
    __tablename__ = 'user'
    id = Column(Integer, primary_key=True)
    username = Column(String(80), unique=True, nullable=False)
    email = Column(String(120), unique=True, nullable=False)

    def __repr__(self):
        return "<User %r>" % self.username


class Index(SessionMixin, tornado.web.RequestHandler):
    def get(self):
        count = self.session.query(User).count()
        if not count:
            user = User()
            user.username = "test"
            user.email = "test@test.com"
            self.session.add(user)
        self.set_header("Content-Type", "text/html")
        self.render("index.html")


class Date(tornado.web.RequestHandler):
    def get(self):
        self.write({"date": "Current Date and Time is: %s" %
                               datetime.now()})
        self.set_header("Content-Type", "application/json")


class Locations(tornado.web.RequestHandler):
    def get(self):
        self.write({"locations": city_list})
        self.set_header("Content-Type", "application/json")


class Weather(tornado.web.RequestHandler):
    def get(self, location_name):
        if location_name not in city_list:
            # self._reason = "Location not found!"
            raise tornado.web.HTTPError(
                status_code=404,
                reason="Location not found!"
            )
        with requests.get(
                'https://airbrake.github.io/weatherapi/weather/' + location_name
        ) as f:
            self.write(f.json())
            self.set_header("Content-Type","application/json")


async def main():
    db.metadata.create_all(bind=db.engine)
    settings = {
        "db": db,
        "PYBRAKE": {
            "project_id": 99999,
            "project_key": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        },
        "template_path": os.path.join(os.path.dirname(__file__), "templates")
    }
    app = tornado.web.Application([
        ("/", Index),
        ("/date", Date),
        ("/locations", Locations),
        (r"/weather/?(?P<location_name>[A-Za-z0-9-]+)?", Weather),
    ], **settings)
    print("Serving on localhost:8000...")
    app = init_app(app)
    app.listen(8000)
    await asyncio.Event().wait()


if __name__ == "__main__":
    asyncio.run(main())



import sys
import time
import traceback

try:
    from bottle import (
        Bottle,
        template,
        request as bottle_request,
        response as bottle_response,
        HTTPError,
        __version__ as BOTTLE_VERSION,
    )
except ImportError as e:
    sys.exit()

from .. import Notifier
from .. import RouteMetric
from ..metrics import (
    set_active as set_active_metrics,
    get_active as get_active_metrics,
    start_span,
    end_span,
)

try:
    from bottle.ext.sqlalchemy import SQLAlchemyPlugin
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True

try:
    from bottle_login import LoginPlugin
except ImportError:
    _bottle_login_available = False
else:
    _bottle_login_available = True

_UNKNOWN_ROUTE = "UNKNOWN"


def request_filter(notice):
    if bottle_request is None:
        return notice
    ctx = notice["context"]
    ctx["method"] = bottle_request.method
    ctx["url"] = bottle_request.url
    ctx["route"] = str(bottle_request.path)

    try:
        user_addr = bottle_request.remote_route[0]
    except IndexError:
        user_addr = bottle_request.environ.get('REMOTE_ADDR')
    if user_addr:
        ctx["userAddr"] = user_addr

    if _bottle_login_available:
        for plugin in bottle_request.app.plugins:
            if isinstance(plugin, LoginPlugin):
                current_user = plugin.get_user()
                user = dict(id=current_user.get_id())
                for s in ["username", "name"]:
                    if hasattr(current_user, s):
                        user[s] = getattr(current_user, s)
                ctx["user"] = user

    notice["params"]["request"] = dict(
        form=dict(bottle_request.forms),
        json=bottle_request.json and dict(bottle_request.json) or {},
        files=dict(bottle_request.files),
        cookies=dict(bottle_request.cookies),
        headers=dict(bottle_request.headers),
        url_rule=bottle_request.url,
        url_args=dict(bottle_request.url_args),
        script_name=bottle_request.script_name,
        query=bottle_request.query,
        query_string=dict(bottle_request.query_string),
    )
    return notice


def before_request_middleware():
    notifier = get_notifier()
    if notifier and not notifier.config.get("performance_stats"):
        return

    if bottle_request.url:
        route = bottle_request.url
    else:
        route = _UNKNOWN_ROUTE

    metric = RouteMetric(method=bottle_request.method, route=route)
    set_active_metrics(metric)


def after_request_middleware():
    notifier = get_notifier()
    if notifier and not notifier.config.get("performance_stats"):
        return bottle_response

    metric = get_active_metrics()
    if metric is not None:
        metric.status_code = bottle_response.status_code
        metric.content_type = bottle_response.content_type
        metric.end_time = time.time()
        notifier.routes.notify(metric)
        set_active_metrics(None)

    return bottle_response


def _handle_exception(app, environ):
    res = app.old_handle(environ)
    if isinstance(res, Exception) and isinstance(res, HTTPError):
        notifier = app.config["pybrake"]
        notice = notifier.build_notice(res)
        notifier.send_notice(notice)
    # scope cleanup
    return res


# Patch for set template render stats
old_template = template


def _template_render(*args, **kwargs):
    start_span("template")
    res = old_template(*args, **kwargs)
    end_span("template")
    return res


template = _template_render


def _sqla_instrument(app):
    try:
        sqla = None
        for plugin in app.plugins:
            if isinstance(plugin, SQLAlchemyPlugin):
                sqla = plugin
                break
        if sqla is None:
            return
    except Exception as err:  # pylint: disable=broad-except
        raise err

    from sqlalchemy import event  # pylint: disable=import-outside-toplevel

    event.listen(sqla.engine, "before_cursor_execute", _before_cursor())
    event.listen(sqla.engine, "after_cursor_execute", _after_cursor())


def _before_cursor():
    def _sqla_before_cursor_execute(
        conn, cursor, statement, parameters, context, executemany
    ):
        notifier = get_notifier()
        if notifier and not notifier.config.get("performance_stats"):
            return
        start_span("sql")

    return _sqla_before_cursor_execute


def _after_cursor():
    def _sqla_after_cursor_execute(
        conn, cursor, statement, parameters, context, executemany
    ):
        notifier = get_notifier()
        if notifier and not notifier.config.get("performance_stats"):
            return
        end_span("sql")
        metric = get_active_metrics()
        if metric is not None:
            try:
                traceback_frm = traceback.extract_stack()[0]
            except IndexError as er:  # pylint: disable=unused-variable
                traceback_frm = None
            notifier.queries.notify(
                query=statement,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=metric.start_time,
                end_time=time.time(),
                function=traceback_frm.name if traceback_frm else '',
                file=traceback_frm.filename if traceback_frm else '',
                line=traceback_frm.lineno if traceback_frm else 0,
            )

    return _sqla_after_cursor_execute


def get_notifier():
    return bottle_request.app.config.get('pybrake')  # pylint: disable=no-member


def init_app(app):
    if "pybrake" in app.config:
        raise ValueError("pybrake is already injected")
    if "PYBRAKE" not in app.config:
        raise ValueError("app.config['PYBRAKE'] is not defined")

    notifier = Notifier(**app.config["PYBRAKE"])

    notifier.add_filter(request_filter)

    app.config["pybrake"] = notifier

    # Patch for handle exception notification
    old_handle = Bottle._handle
    Bottle.old_handle = old_handle
    Bottle._handle = _handle_exception

    app.add_hook('before_request', before_request_middleware)
    app.add_hook('after_request', after_request_middleware)

    if _sqla_available:
        _sqla_instrument(app)

    return app



import base64
import json
from threading import Lock, Timer

from . import metrics
from . import constant
from .backlog import Backlog
from .route_metric import RouteBreakdowns
from .tdigest import as_bytes, TDigestStat
from .utils import time_trunc_minute


class _Routes:
    """
    Routes and route breakdown stats are transmitted to the Airbrake site
    using _Routes. When the notifier is initialized, it will initialize the
    RouteStats and RouteBreakdowns instances.
    """

    def __init__(self, **kwargs):
        self.config = kwargs["config"]
        self.stats = RouteStats(**kwargs)
        self.breakdowns = RouteBreakdowns(**kwargs)

    def notify(self, metric):
        if not self.config.get("performance_stats"):
            return

        metric.end()
        self.stats.notify(metric)
        self.breakdowns.notify(metric)


class RouteStat(TDigestStat):
    """
    RouteStat will collect request execution statistics such as request start
    time, end time, request endpoint, response status code, and route method.
    """

    def __new__(cls, *, method="", route="", status_code=0, time):
        instance = super(RouteStat, cls).__new__(cls)
        instance.__slots__ = instance.__slots__ + (
            "method", "route", "statusCode", "time"
        )
        return instance

    @property
    def __dict__(self):
        b = as_bytes(self.td)
        self.tdigest = base64.b64encode(b).decode("ascii")
        return {s: getattr(self, s) for s in self.__slots__ if s != "td"}

    def __init__(self, *, method="", route="", status_code=0, time=None):
        super().__init__()
        self.method = method
        self.route = route
        self.statusCode = status_code
        self.time = time_trunc_minute(time)


class RouteStats:
    def __init__(self, *, project_id=0, project_key="", **kwargs):
        self._config = kwargs["config"]

        self._project_id = project_id
        self._ab_headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + project_key,
        }
        self._env = kwargs.get("environment")

        self._thread = None
        self._lock = Lock()
        self._stats = None
        self._backlog = None
        if self._config.get('backlog_enabled'):
            if not metrics.APM_Backlog:
                metrics.APM_Backlog = Backlog(
                    header=self._ab_headers,
                    method="POST",
                    maxlen=self._config.get('max_backlog_size'),
                )
            self._backlog = metrics.APM_Backlog

    def notify(self, metric):
        if not self._config.get("performance_stats"):
            return

        key = route_stat_key(
            method=metric.method,
            route=metric.route,
            status_code=metric.status_code,
            time=metric.start_time,
        )
        with self._lock:
            if self._stats is None:
                self._stats = {}
                self._thread = Timer(constant.FLUSH_PERIOD, self._flush)
                self._thread.start()

            if key in self._stats:
                stat = self._stats[key]
            else:
                stat = RouteStat(
                    method=metric.method,
                    route=metric.route,
                    status_code=metric.status_code,
                    time=metric.start_time,
                )
                self._stats[key] = stat

            ms = (metric.end_time - metric.start_time) * 1000
            stat.add(ms)

    def _flush(self):
        """
        TODO: Below disabled pylint will remove in refactoring
        """
        # pylint: disable=too-many-branches
        stats = None
        with self._lock:
            stats = self._stats
            self._stats = None

        if not stats:
            raise ValueError("stats is empty")

        out = {"routes": [v.__dict__ for v in stats.values()]}
        if self._env:
            out["environment"] = self._env

        out = json.dumps(out).encode("utf8")
        metrics.send(
            url=self._ab_url(), headers=self._ab_headers,
            method="POST", payload=out,
        )

    def _ab_url(self):
        return f"{self._config.get('apm_host')}/api/v5/projects" \
               f"/{self._project_id}/routes-stats"


def route_stat_key(*, method="", route="", status_code=0, time=None):
    time = time // 60 * 60
    return method, route, status_code, time



from .starlette import init_pybrake


def init_app(app, sqlEngine=None):
    """
    Initiate the pybrake notifier and apply the patch for
    error monitoring and APM.
    :param app: instance of FastAPI application
    :param sqlEngine: SQLALCHEMY engine instance
    :return: FastAPI application instance after apply patch or new setting
    """
    if "pybrake" in app.extra:
        raise ValueError("pybrake is already injected")
    if "PYBRAKE" not in app.extra:
        raise ValueError("app.config['PYBRAKE'] is not defined")

    app, notifier = init_pybrake(app, app.extra["PYBRAKE"], sqlEngine)
    app.extra["pybrake"] = notifier
    return app



from datetime import datetime

import requests
from fastapi import FastAPI
from fastapi.responses import HTMLResponse
from pybrake.middleware.fastapi import init_app
from sqlalchemy import Column, Integer, String
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from starlette.responses import JSONResponse

app = FastAPI(debug=True)
app.extra["PYBRAKE"] = dict(
    project_id=999999,
    project_key='xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx',
    environment="test",
    performance_stats=True,  # False to disable APM
)

# SQL setup
SQLALCHEMY_DATABASE_URL = "sqlite:///./sql_app.db"

engine = create_engine(
    SQLALCHEMY_DATABASE_URL, connect_args={"check_same_thread": False}
)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

Base = declarative_base()

db = SessionLocal()


class User(Base):
    __tablename__ = 'user'
    id = Column(Integer, primary_key=True)
    username = Column(String(80), unique=True, nullable=False)
    email = Column(String(120), unique=True, nullable=False)

    def __repr__(self):
        return "<User %r>" % self.username


Base.metadata.create_all(bind=engine)

app = init_app(app, engine)

city_list = ["pune", "austin", "santabarbara", "washington"]


# API for Hello Application
@app.get("/", response_class=HTMLResponse)
async def root():
    print(db.query(User).all())
    html_content = """
        <html>
            <head>
                <title>Welcome to the weather app.</title>
            </head>
            <body>
                <h1>Hello, Welcome to the weather app.</h1>
            </body>
        </html>
        """
    return HTMLResponse(content=html_content, status_code=200)


# API for current server date
@app.get("/date")
async def date():
    current_datetime = datetime.now()
    return {"Current date and time": current_datetime}


# API for location details
@app.get("/locations")
async def locations():
    return {"locations list": city_list}


# API for weather details for a location
@app.get("/weather/{location_name}")
async def weather(location_name):
    if location_name not in city_list:
        return JSONResponse(content={
            "message": "Location not found!"
        }, status_code=404)
    with requests.get(
            'https://airbrake.github.io/weatherapi/weather/' + location_name
    ) as f:
        data = f.json()
        return data



from masonite.environment import env


DRIVERS = {
    "default": env("QUEUE_DRIVER", "async"),
    "database": {
        "connection": "sqlite",
        "table": "jobs",
        "failed_table": "failed_jobs",
        "attempts": 3,
        "poll": 5,
    },
    "redis": {
        "name": env("QUEUE_USERNAME", "guest"),
        "password": env("QUEUE_PASSWORD", "guest"),
        "port": env("QUEUE_PORT", "6379"),
        "host": env("QUEUE_HOST", "localhost"),
    },
    "amqp": {
        "username": env("QUEUE_USERNAME", "guest"),
        "password": env("QUEUE_PASSWORD", "guest"),
        "port": env("QUEUE_PORT", "5672"),
        "vhost": env("QUEUE_VHOST", ""),
        "host": env("QUEUE_HOST", "localhost"),
        "exchange": "",
        # See https://pika.readthedocs.io/en/stable/modules/parameters.html#pika.connection.URLParameters
        # for valid connection options values
        "connection_options": {},
        "channel": env("QUEUE_CHANNEL", "default"),
        "queue": "masonite4",
        "tz": "UTC",
    },
    "async": {
        "blocking": True,
        "callback": "handle",
        "mode": "threading",
        "workers": 1,
    },
}



# Notifier Name
notifier_name = "pybrake"

# Pybrake Notifier Version
version = "1.10.1"

# End point of Airbrake API host. That will always be as default value.
AIRBRAKE_HOST = "https://api.airbrake.io"

# End point of Airbrake remote configuration.
AIRBRAKE_CONFIG_HOST = "https://notifier-configs.airbrake.io"

FLUSH_PERIOD = 15
MaxRetryAttempt = 1

HTTP_HANDLER = "http.handler"

QUEUE_HANDLER = "queue.handler"
